#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# Python standard library
import time
import wrapt
import queue
import struct
import logging
import binascii

from nordicsemi.dfu.dfu_transport   import DfuTransport, DfuEvent
from pc_ble_driver_py.exceptions    import NordicSemiException, IllegalStateException
from pc_ble_driver_py.ble_driver    import BLEDriver, BLEDriverObserver, BLEEnableParams, BLEUUIDBase, BLEGapSecKDist, BLEGapSecParams, \
    BLEGapIOCaps, BLEUUID, BLEAdvData, BLEGapConnParams, NordicSemiErrorCheck, BLEGapSecStatus, driver
from pc_ble_driver_py.ble_driver    import ATT_MTU_DEFAULT, BLEConfig, BLEConfigConnGatt, BLEConfigConnGap
from pc_ble_driver_py.ble_adapter   import BLEAdapter, BLEAdapterObserver, EvtSync

logger  = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

from pc_ble_driver_py import config
global nrf_sd_ble_api_ver
nrf_sd_ble_api_ver = config.sd_api_ver_get()


class ValidationException(NordicSemiException):
    """"
    Exception used when validation failed
    """
    pass


class DFUAdapter(BLEDriverObserver, BLEAdapterObserver):

    BASE_UUID = BLEUUIDBase([0x8E, 0xC9, 0x00, 0x00, 0xF3, 0x15, 0x4F, 0x60,
                             0x9F, 0xB8, 0x83, 0x88, 0x30, 0xDA, 0xEA, 0x50])

    # Buttonless characteristics
    BLE_DFU_BUTTONLESS_CHAR_UUID        = BLEUUID(0x0003, BASE_UUID)
    BLE_DFU_BUTTONLESS_BONDED_CHAR_UUID = BLEUUID(0x0004, BASE_UUID)
    SERVICE_CHANGED_UUID                = BLEUUID(0x2A05)

    # Bootloader characteristics
    CP_UUID     = BLEUUID(0x0001, BASE_UUID)
    DP_UUID     = BLEUUID(0x0002, BASE_UUID)

    CONNECTION_ATTEMPTS   = 3
    ERROR_CODE_POS        = 2
    LOCAL_ATT_MTU         = 247

    def __init__(self, adapter, bonded=False, keyset=None):
        super().__init__()

        self.evt_sync           = EvtSync(['connected', 'disconnected', 'sec_params',
                                           'auth_status', 'conn_sec_update'])
        self.conn_handle        = None
        self.adapter            = adapter
        self.bonded             = bonded
        self.keyset             = keyset
        self.notifications_q    = queue.Queue()
        self.indication_q       = queue.Queue()
        self.att_mtu            = ATT_MTU_DEFAULT
        self.packet_size        = self.att_mtu - 3
        self.adapter.observer_register(self)
        self.adapter.driver.observer_register(self)

    def open(self):
        self.adapter.driver.open()

        assert nrf_sd_ble_api_ver in [2, 5]

        if nrf_sd_ble_api_ver == 2:
            self.adapter.driver.ble_enable(
                BLEEnableParams(
                    vs_uuid_count = 10,
                    service_changed = True,
                    periph_conn_count = 0,
                    central_conn_count = 1,
                    central_sec_count = 1,
                )
            )

        if nrf_sd_ble_api_ver == 5:
            self.adapter.driver.ble_cfg_set(
                BLEConfig.conn_gatt,
                BLEConfigConnGatt(att_mtu=DFUAdapter.LOCAL_ATT_MTU),
            )
            self.adapter.driver.ble_cfg_set(
                BLEConfig.conn_gap,
                BLEConfigConnGap(event_length=5))  # Event length 5 is required for max data length
            self.adapter.driver.ble_enable()

        self.adapter.driver.ble_vs_uuid_add(DFUAdapter.BASE_UUID)

    def close(self):
        if self.conn_handle is not None:
            logger.info('BLE: Disconnecting from target')
            self.adapter.disconnect(self.conn_handle)
            self.evt_sync.wait('disconnected')
        self.conn_handle    = None
        self.evt_sync       = None
        self.adapter.observer_unregister(self)
        self.adapter.driver.observer_unregister(self)
        self.adapter.driver.close()

    def connect(self, target_device_name, target_device_addr):
        """ Connect to Bootloader or Application with Buttonless Service.

        Args:
            target_device_name (str): Device name to scan for.
            target_device_addr (str): Device addr to scan for.
        """
        self.target_device_name = target_device_name
        self.target_device_addr = target_device_addr

        logger.info('BLE: Scanning for {}'.format(self.target_device_name))
        self.adapter.driver.ble_gap_scan_start()
        self.verify_stable_connection()
        if self.conn_handle is None:
            raise NordicSemiException('Timeout. Target device not found.')

        logger.info('BLE: Service Discovery')
        self.adapter.service_discovery(conn_handle=self.conn_handle)

        # Check if connected peer has Buttonless service.
        if self.adapter.db_conns[self.conn_handle].get_cccd_handle(DFUAdapter.BLE_DFU_BUTTONLESS_CHAR_UUID):
            self.jump_from_buttonless_mode_to_bootloader(DFUAdapter.BLE_DFU_BUTTONLESS_CHAR_UUID)
        elif self.adapter.db_conns[self.conn_handle].get_cccd_handle(DFUAdapter.BLE_DFU_BUTTONLESS_BONDED_CHAR_UUID):
            self.jump_from_buttonless_mode_to_bootloader(DFUAdapter.BLE_DFU_BUTTONLESS_BONDED_CHAR_UUID)

        if self.bonded:
            # For combined Updates with bonds enabled, re-encryption is needed
            self.encrypt()

        if nrf_sd_ble_api_ver >= 3:
            if DFUAdapter.LOCAL_ATT_MTU > ATT_MTU_DEFAULT:
                logger.info('BLE: Enabling longer ATT MTUs')
                self.att_mtu = self.adapter.att_mtu_exchange(self.conn_handle, DFUAdapter.LOCAL_ATT_MTU)

                logger.info('BLE: Enabling longer Data Length')
                max_data_length = 251  # Max data length for SD v5
                data_length = self.att_mtu + 4  # ATT PDU overhead is 4
                if data_length > max_data_length:
                    data_length = max_data_length
                self.adapter.data_length_update(self.conn_handle, data_length)
            else:
                logger.info('BLE: Using default ATT MTU')

        logger.debug('BLE: Enabling Notifications')
        self.adapter.enable_notification(conn_handle=self.conn_handle, uuid=DFUAdapter.CP_UUID)
        return self.target_device_name, self.target_device_addr

    def jump_from_buttonless_mode_to_bootloader(self, buttonless_uuid):
        """ Function for going to bootloader mode from application with
         buttonless service. It supports both bonded and unbonded
         buttonless characteristics.

        Args:
            buttonless_uuid: UUID of discovered buttonless characteristic.

        """
        if buttonless_uuid == DFUAdapter.BLE_DFU_BUTTONLESS_BONDED_CHAR_UUID:
            logger.info("Bonded Buttonless characteristic discovered -> Bond")
            self.bond()
        else:
            logger.info("Un-bonded Buttonless characteristic discovered -> Increment target device addr")
            self.target_device_addr = "{:X}".format(int(self.target_device_addr, 16) + 1)
            self.target_device_addr_type.addr[-1] += 1

        # Enable indication for Buttonless DFU Service
        self.adapter.enable_indication(self.conn_handle, buttonless_uuid)

        # Enable indication for Service changed Service, if present.
        if self.adapter.db_conns[self.conn_handle].get_char_handle(DFUAdapter.SERVICE_CHANGED_UUID):
            self.adapter.enable_indication(self.conn_handle, DFUAdapter.SERVICE_CHANGED_UUID)

        # Enter DFU mode
        self.adapter.write_req(self.conn_handle, buttonless_uuid, [0x01])
        response = self.indication_q.get(timeout=DfuTransportBle.DEFAULT_TIMEOUT)
        if response[DFUAdapter.ERROR_CODE_POS] != 0x01:
            raise Exception("Error - Unexpected response")

        # Wait for buttonless peer to disconnect
        self.evt_sync.wait('disconnected')

        # Reconnect
        self.target_device_name = None
        self.adapter.driver.ble_gap_scan_start()
        self.verify_stable_connection()
        if self.conn_handle is None:
            raise NordicSemiException('Timeout. Target device not found.')
        logger.info('BLE: Connected to target')

        logger.debug('BLE: Service Discovery')
        self.adapter.service_discovery(conn_handle=self.conn_handle)

    def verify_stable_connection(self):
        """ Verify connection event, and verify that unexpected disconnect
         events are not received.

        Returns:
            True if connected, else False.

        """
        self.conn_handle = self.evt_sync.wait('connected')
        if self.conn_handle is not None:
            retries = DFUAdapter.CONNECTION_ATTEMPTS
            while retries:
                if self.evt_sync.wait('disconnected', timeout=1) is None:
                    break

                logger.warning("Received unexpected disconnect event, "
                               "trying to re-connect to: {}".format(self.target_device_addr))
                time.sleep(1)

                self.adapter.connect(address=self.target_device_addr_type,
                                     conn_params=self.conn_params,
                                     tag=1)
                self.conn_handle = self.evt_sync.wait('connected')
                retries -= 1
            else:
                if self.evt_sync.wait('disconnected', timeout=1) is not None:
                    raise Exception("Failure - Connection failed due to 0x3e")

            logger.info("Successfully Connected")
            return

        self.adapter.driver.ble_gap_scan_stop()
        raise Exception("Connection Failure - Device not found!")

    def setup_keyset(self):
        """ Setup keyset structure.

        """
        self.keyset = driver.ble_gap_sec_keyset_t()

        self.id_key_own = driver.ble_gap_id_key_t()
        self.id_key_peer = driver.ble_gap_id_key_t()

        self.enc_key_own = driver.ble_gap_enc_key_t()
        self.enc_key_peer = driver.ble_gap_enc_key_t()

        self.sign_info_own = driver.ble_gap_sign_info_t()
        self.sign_info_peer = driver.ble_gap_sign_info_t()

        self.lesc_pk_own = driver.ble_gap_lesc_p256_pk_t()
        self.lesc_pk_peer = driver.ble_gap_lesc_p256_pk_t()

        self.keyset.keys_own.p_enc_key   = self.enc_key_own
        self.keyset.keys_own.p_id_key    = self.id_key_own
        self.keyset.keys_own.p_sign_key  = self.sign_info_own
        self.keyset.keys_own.p_pk        = self.lesc_pk_own
        self.keyset.keys_peer.p_enc_key  = self.enc_key_peer
        self.keyset.keys_peer.p_id_key   = self.id_key_peer
        self.keyset.keys_peer.p_sign_key = self.sign_info_peer
        self.keyset.keys_peer.p_pk       = self.lesc_pk_peer

    def setup_sec_params(self):
        """ Setup Security parameters.

        """

        self.kdist_own = BLEGapSecKDist(enc=True,
                                        id=True,
                                        sign=False,
                                        link=False)
        self.kdist_peer = BLEGapSecKDist(enc=True,
                                         id=True,
                                         sign=False,
                                         link=False)
        self.sec_params = BLEGapSecParams(bond=True,
                                          mitm=False,
                                          lesc=False,
                                          keypress=False,
                                          io_caps=BLEGapIOCaps.none,
                                          oob=False,
                                          min_key_size=7,
                                          max_key_size=16,
                                          kdist_own=self.kdist_own,
                                          kdist_peer=self.kdist_peer)

    def bond(self):
        """ Bond to Application with Buttonless Service.

        """
        self.bonded = True
        self.setup_sec_params()
        self.setup_keyset()

        self.adapter.driver.ble_gap_authenticate(self.conn_handle, self.sec_params)
        self.evt_sync.wait(evt="sec_params")
        self.adapter.driver.ble_gap_sec_params_reply(self.conn_handle,
                                                     BLEGapSecStatus.success,
                                                     None,
                                                     self.keyset,
                                                     None)

        result = self.evt_sync.wait(evt="auth_status")
        if result != BLEGapSecStatus.success:
            raise NordicSemiException("Auth Status returned error code: {}".format(result))

    def encrypt(self):
        """ Re-encrypt to bootloader.

        """
        logger.info("Re-encryption to bootloader")
        self.adapter.driver.ble_gap_encrypt(self.conn_handle,
                                            self.keyset.keys_peer.p_enc_key.master_id,
                                            self.keyset.keys_peer.p_enc_key.enc_info)
        self.evt_sync.wait('conn_sec_update')

    def write_control_point(self, data):
        self.adapter.write_req(self.conn_handle, DFUAdapter.CP_UUID, data)

    def write_data_point(self, data):
        self.adapter.write_cmd(self.conn_handle, DFUAdapter.DP_UUID, data)

    def on_gap_evt_sec_params_request(self, ble_driver, conn_handle, peer_params):
        logger.info("Got sec params req")
        self.evt_sync.notify(evt='sec_params', data=conn_handle)

    def on_gap_evt_auth_status(self, ble_driver, conn_handle, auth_status):
        logger.info("Got auth status:{}".format(auth_status))
        self.evt_sync.notify(evt='auth_status', data=auth_status)

    def on_gap_evt_conn_sec_update(self, ble_driver, conn_handle):
        logger.info("Got Conn sec update")
        self.evt_sync.notify(evt='conn_sec_update', data=conn_handle)

    def on_gap_evt_connected(self, ble_driver, conn_handle, peer_addr, role, conn_params):
        self.evt_sync.notify(evt = 'connected', data = conn_handle)
        logger.info('BLE: Connected to {}'.format(peer_addr.addr))

    def on_gap_evt_disconnected(self, ble_driver, conn_handle, reason):
        self.evt_sync.notify(evt = 'disconnected', data = conn_handle)
        self.conn_handle = None
        logger.info('BLE: Disconnected with reason: {}'.format(reason))

    def on_gap_evt_adv_report(self, ble_driver, conn_handle, peer_addr, rssi, adv_type, adv_data):
        dev_name_list = []
        if BLEAdvData.Types.complete_local_name in adv_data.records:
            dev_name_list = adv_data.records[BLEAdvData.Types.complete_local_name]

        elif BLEAdvData.Types.short_local_name in adv_data.records:
            dev_name_list = adv_data.records[BLEAdvData.Types.short_local_name]

        dev_name        = "".join(chr(e) for e in dev_name_list)
        address_string  = "".join("{0:02X}".format(b) for b in peer_addr.addr)
        logger.info('Received advertisement report, address: 0x{}, device_name: {}'.format(address_string, dev_name))

        if (dev_name == self.target_device_name) or (address_string == self.target_device_addr):
            self.conn_params = BLEGapConnParams(min_conn_interval_ms = 7.5,
                                                max_conn_interval_ms = 30,
                                                conn_sup_timeout_ms  = 4000,
                                                slave_latency        = 0)
            logger.info('BLE: Found target advertiser, address: 0x{}, name: {}'.format(address_string, dev_name))
            logger.info('BLE: Connecting to 0x{}'.format(address_string))
            # Connect must specify tag=1 to enable the settings
            # set with BLEConfigConnGatt (that implicitly operates
            # on connections with tag 1) to allow for larger MTU.
            self.adapter.connect(address=peer_addr,
                                 conn_params=self.conn_params,
                                 tag=1)
            # store the address for subsequent connections
            self.target_device_addr = address_string
            self.target_device_addr_type = peer_addr

    def on_notification(self, ble_adapter, conn_handle, uuid, data):
        if self.conn_handle         != conn_handle: return
        if DFUAdapter.CP_UUID.value != uuid.value:
            return
        self.notifications_q.put(data)

    def on_indication(self, ble_adapter, conn_handle, uuid, data):
        if self.conn_handle         != conn_handle: return
        if DFUAdapter.BLE_DFU_BUTTONLESS_BONDED_CHAR_UUID.value != uuid.value and \
           DFUAdapter.BLE_DFU_BUTTONLESS_CHAR_UUID.value != uuid.value:
            return
        self.indication_q.put(data)

    def on_gattc_evt_exchange_mtu_rsp(self, ble_driver, conn_handle, *, status, att_mtu):
        logger.info('ATT MTU exchanged: conn_handle={} att_mtu={}'.format(conn_handle, att_mtu))
        self.att_mtu = att_mtu
        self.packet_size = att_mtu - 3


class DfuBLEDriver(BLEDriver):
    def __init__(self, serial_port, baud_rate=115200, auto_flash=False):
        super().__init__(serial_port, baud_rate)

    @NordicSemiErrorCheck
    @wrapt.synchronized(BLEDriver.api_lock)
    def ble_gap_sec_params_reply(self, conn_handle, sec_status, sec_params, own_keys, peer_keys):
        assert isinstance(sec_status, BLEGapSecStatus), 'Invalid argument type'
        assert sec_params is None or isinstance(sec_params, BLEGapSecParams), 'Invalid argument type'
        assert peer_keys is None, 'NOT IMPLEMENTED'

        return driver.sd_ble_gap_sec_params_reply(self.rpc_adapter,
                                                  conn_handle,
                                                  sec_status.value,
                                                  sec_params.to_c() if sec_params else None,
                                                  own_keys)


class DfuTransportBle(DfuTransport):

    DEFAULT_TIMEOUT     = 20
    RETRIES_NUMBER      = 3

    def __init__(self,
                 serial_port,
                 att_mtu,
                 target_device_name=None,
                 target_device_addr=None,
                 baud_rate=1000000,
                 prn=0):
        super().__init__()
        DFUAdapter.LOCAL_ATT_MTU = att_mtu
        self.baud_rate          = baud_rate
        self.serial_port        = serial_port
        self.att_mtu            = att_mtu
        self.target_device_name = target_device_name
        self.target_device_addr = target_device_addr
        self.dfu_adapter        = None
        self.prn                = prn

        self.bonded             = False
        self.keyset             = None

    def open(self):
        if self.dfu_adapter:
            raise IllegalStateException('DFU Adapter is already open')

        super().open()
        driver           = DfuBLEDriver(serial_port = self.serial_port,
                                        baud_rate   = self.baud_rate)
        adapter          = BLEAdapter(driver)
        self.dfu_adapter = DFUAdapter(adapter=adapter, bonded=self.bonded, keyset=self.keyset)
        self.dfu_adapter.open()
        self.target_device_name, self.target_device_addr = self.dfu_adapter.connect(
                                                        target_device_name = self.target_device_name,
                                                        target_device_addr = self.target_device_addr)
        self.__set_prn()

    def close(self):

        # Get bonded status and BLE keyset from DfuAdapter
        self.bonded = self.dfu_adapter.bonded
        self.keyset = self.dfu_adapter.keyset

        if not self.dfu_adapter:
            raise IllegalStateException('DFU Adapter is already closed')
        super().close()
        self.dfu_adapter.close()
        self.dfu_adapter = None

    def send_init_packet(self, init_packet):
        def try_to_recover():
            if response['offset'] == 0 or response['offset'] > len(init_packet):
                # There is no init packet or present init packet is too long.
                return False

            expected_crc = (binascii.crc32(init_packet[:response['offset']]) & 0xFFFFFFFF)

            if expected_crc != response['crc']:
                # Present init packet is invalid.
                return False

            if len(init_packet) > response['offset']:
                # Send missing part.
                try:
                    self.__stream_data(data     = init_packet[response['offset']:],
                                       crc      = expected_crc,
                                       offset   = response['offset'])
                except ValidationException:
                    return False

            self.__execute()
            return True

        response = self.__select_command()
        assert len(init_packet) <= response['max_size'], 'Init command is too long'

        if try_to_recover():
            return

        for r in range(DfuTransportBle.RETRIES_NUMBER):
            try:
                self.__create_command(len(init_packet))
                self.__stream_data(data=init_packet)
                self.__execute()
            except ValidationException:
                pass
            break
        else:
            raise NordicSemiException("Failed to send init packet")

    def send_firmware(self, firmware):
        def try_to_recover():
            if response['offset'] == 0:
                # Nothing to recover
                return

            expected_crc = binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
            remainder    = response['offset'] % response['max_size']

            if expected_crc != response['crc']:
                # Invalid CRC. Remove corrupted data.
                response['offset'] -= remainder if remainder != 0 else response['max_size']
                response['crc']     = binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
                return

            if (remainder != 0) and (response['offset'] != len(firmware)):
                # Send rest of the page.
                try:
                    to_send             = firmware[response['offset'] : response['offset'] + response['max_size'] - remainder]
                    response['crc']     = self.__stream_data(data   = to_send,
                                                             crc    = response['crc'],
                                                             offset = response['offset'])
                    response['offset'] += len(to_send)
                except ValidationException:
                    # Remove corrupted data.
                    response['offset'] -= remainder
                    response['crc']     = binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
                    return

            self.__execute()
            self._send_event(event_type=DfuEvent.PROGRESS_EVENT, progress=response['offset'])

        response = self.__select_data()
        try_to_recover()

        for i in range(response['offset'], len(firmware), response['max_size']):
            data = firmware[i:i+response['max_size']]
            for r in range(DfuTransportBle.RETRIES_NUMBER):
                try:
                    self.__create_data(len(data))
                    response['crc'] = self.__stream_data(data=data, crc=response['crc'], offset=i)
                    self.__execute()
                except ValidationException:
                    pass
                break
            else:
                raise NordicSemiException("Failed to send firmware")
            self._send_event(event_type=DfuEvent.PROGRESS_EVENT, progress=len(data))

    def __set_prn(self):
        logger.debug("BLE: Set Packet Receipt Notification {}".format(self.prn))
        self.dfu_adapter.write_control_point([DfuTransportBle.OP_CODE['SetPRN']] + list(struct.pack('<H', self.prn)))
        self.__get_response(DfuTransportBle.OP_CODE['SetPRN'])

    def __create_command(self, size):
        self.__create_object(0x01, size)

    def __create_data(self, size):
        self.__create_object(0x02, size)

    def __create_object(self, object_type, size):
        self.dfu_adapter.write_control_point([DfuTransportBle.OP_CODE['CreateObject'], object_type]\
                                            + list(struct.pack('<L', size)))
        self.__get_response(DfuTransportBle.OP_CODE['CreateObject'])

    def __calculate_checksum(self):
        self.dfu_adapter.write_control_point([DfuTransportBle.OP_CODE['CalcChecSum']])
        response = self.__get_response(DfuTransportBle.OP_CODE['CalcChecSum'])

        (offset, crc) = struct.unpack('<II', bytearray(response))
        return {'offset': offset, 'crc': crc}

    def __execute(self):
        self.dfu_adapter.write_control_point([DfuTransportBle.OP_CODE['Execute']])
        self.__get_response(DfuTransportBle.OP_CODE['Execute'])

    def __select_command(self):
        return self.__select_object(0x01)

    def __select_data(self):
        return self.__select_object(0x02)

    def __select_object(self, object_type):
        logger.debug("BLE: Selecting Object: type:{}".format(object_type))
        self.dfu_adapter.write_control_point([DfuTransportBle.OP_CODE['ReadObject'], object_type])
        response = self.__get_response(DfuTransportBle.OP_CODE['ReadObject'])

        (max_size, offset, crc)= struct.unpack('<III', bytearray(response))
        logger.debug("BLE: Object selected: max_size:{} offset:{} crc:{}".format(max_size, offset, crc))
        return {'max_size': max_size, 'offset': offset, 'crc': crc}

    def __get_checksum_response(self):
        response = self.__get_response(DfuTransportBle.OP_CODE['CalcChecSum'])

        (offset, crc) = struct.unpack('<II', bytearray(response))
        return {'offset': offset, 'crc': crc}

    def __stream_data(self, data, crc=0, offset=0):
        logger.debug("BLE: Streaming Data: len:{0} offset:{1} crc:0x{2:08X}".format(len(data), offset, crc))
        def validate_crc():
            if (crc != response['crc']):
                raise ValidationException('Failed CRC validation.\n'\
                                + 'Expected: {} Received: {}.'.format(crc, response['crc']))
            if (offset != response['offset']):
                raise ValidationException('Failed offset validation.\n'\
                                + 'Expected: {} Received: {}.'.format(offset, response['offset']))

        current_pnr = 0
        for i in range(0, len(data), self.dfu_adapter.packet_size):
            to_transmit     = data[i:i + self.dfu_adapter.packet_size]
            self.dfu_adapter.write_data_point(list(to_transmit))
            crc     = binascii.crc32(to_transmit, crc) & 0xFFFFFFFF
            offset += len(to_transmit)
            current_pnr    += 1
            if self.prn == current_pnr:
                current_pnr = 0
                response    = self.__get_checksum_response()
                validate_crc()

        response = self.__calculate_checksum()
        validate_crc()

        return crc

    def __get_response(self, operation):
        def get_dict_key(dictionary, value):
            return next((key for key, val in list(dictionary.items()) if val == value), None)

        try:
            resp = self.dfu_adapter.notifications_q.get(timeout=DfuTransportBle.DEFAULT_TIMEOUT)
        except queue.Empty:
            raise NordicSemiException('Timeout: operation - {}'.format(get_dict_key(DfuTransportBle.OP_CODE,
                                                                                    operation)))

        if resp[0] != DfuTransportBle.OP_CODE['Response']:
            raise NordicSemiException('No Response: 0x{:02X}'.format(resp[0]))

        if resp[1] != operation:
            raise NordicSemiException('Unexpected Executed OP_CODE.\n' \
                                    + 'Expected: 0x{:02X} Received: 0x{:02X}'.format(operation, resp[1]))

        if resp[2] == DfuTransport.RES_CODE['Success']:
            return resp[3:]

        elif resp[2] == DfuTransport.RES_CODE['ExtendedError']:
            try:
                data = DfuTransport.EXT_ERROR_CODE[resp[3]]
            except IndexError:
                data = "Unsupported extended error type {}".format(resp[3])
            raise NordicSemiException('Extended Error 0x{:02X}: {}'.format(resp[3], data))
        else:
            raise NordicSemiException('Response Code {}'.format(get_dict_key(DfuTransport.RES_CODE, resp[2])))



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# Python standard library
import os
import time
import shutil
import logging
import tempfile


# Nordic libraries
from nordicsemi.dfu.package         import Package

logger = logging.getLogger(__name__)


class Dfu:
    """ Class to handle upload of a new hex image to the device. """

    def __init__(self, zip_file_path, dfu_transport, connect_delay):
        """
        Initializes the dfu upgrade, unpacks zip and registers callbacks.

        @param zip_file_path: Path to the zip file with the firmware to upgrade
        @type zip_file_path: str
        @param dfu_transport: Transport backend to use to upgrade
        @type dfu_transport: nordicsemi.dfu.dfu_transport.DfuTransport
        @param connect_delay: Delay in seconds before each connection to the DFU target
        @type connect_delay: int
        @return
        """
        self.temp_dir           = tempfile.mkdtemp(prefix="nrf_dfu_")
        self.unpacked_zip_path  = os.path.join(self.temp_dir, 'unpacked_zip')
        self.manifest           = Package.unpack_package(zip_file_path, self.unpacked_zip_path)

        self.dfu_transport      = dfu_transport

        if connect_delay is not None:
            self.connect_delay = connect_delay
        else:
            self.connect_delay = 3

    def __del__(self):
        """
        Destructor removes the temporary directory for the unpacked zip
        :return:
        """
        shutil.rmtree(self.temp_dir)


    def _dfu_send_image(self, firmware):
        time.sleep(self.connect_delay)
        self.dfu_transport.open()

        start_time = time.time()

        logger.info("Sending init packet...")
        with open(os.path.join(self.unpacked_zip_path, firmware.dat_file), 'rb') as f:
            data    = f.read()
            self.dfu_transport.send_init_packet(data)

        logger.info("Sending firmware file...")
        with open(os.path.join(self.unpacked_zip_path, firmware.bin_file), 'rb') as f:
            data    = f.read()
            self.dfu_transport.send_firmware(data)

        end_time = time.time()
        logger.info("Image sent in {0}s".format(end_time - start_time))

        self.dfu_transport.close()


    def dfu_send_images(self):
        """
        Does DFU for all firmware images in the stored manifest.
        :return:
        """
        if self.manifest.softdevice_bootloader:
            logger.info("Sending SoftDevice+Bootloader image.")
            self._dfu_send_image(self.manifest.softdevice_bootloader)

        if self.manifest.softdevice:
            logger.info("Sending SoftDevice image...")
            self._dfu_send_image(self.manifest.softdevice)

        if self.manifest.bootloader:
            logger.info("Sending Bootloader image.")
            self._dfu_send_image(self.manifest.bootloader)

        if self.manifest.application:
            logger.info("Sending Application image.")
            self._dfu_send_image(self.manifest.application)


    def dfu_get_total_size(self):
        total_size = 0

        if self.manifest.softdevice_bootloader:
            total_size += os.path.getsize(os.path.join(self.unpacked_zip_path,
                                                       self.manifest.softdevice_bootloader.bin_file))

        if self.manifest.softdevice:
            total_size += os.path.getsize(os.path.join(self.unpacked_zip_path,
                                                       self.manifest.softdevice.bin_file))

        if self.manifest.bootloader:
            total_size += os.path.getsize(os.path.join(self.unpacked_zip_path,
                                                       self.manifest.bootloader.bin_file))

        if self.manifest.application:
            total_size += os.path.getsize(os.path.join(self.unpacked_zip_path,
                                                       self.manifest.application.bin_file))

        return total_size



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# Python imports
import time
from datetime import datetime, timedelta
import binascii
import logging
import struct

# Python 3rd party imports
from serial import Serial
from serial.serialutil import SerialException

# Nordic Semiconductor imports
from nordicsemi.dfu.dfu_transport   import DfuTransport, DfuEvent, TRANSPORT_LOGGING_LEVEL
from pc_ble_driver_py.exceptions    import NordicSemiException
from nordicsemi.lister.device_lister import DeviceLister
from nordicsemi.dfu.dfu_trigger import DFUTrigger

class ValidationException(NordicSemiException):
    """"
    Exception used when validation failed
    """
    pass


logger = logging.getLogger(__name__)

class Slip:
    SLIP_BYTE_END             = 0o300
    SLIP_BYTE_ESC             = 0o333
    SLIP_BYTE_ESC_END         = 0o334
    SLIP_BYTE_ESC_ESC         = 0o335

    SLIP_STATE_DECODING                 = 1
    SLIP_STATE_ESC_RECEIVED             = 2
    SLIP_STATE_CLEARING_INVALID_PACKET  = 3

    @staticmethod
    def encode(data):
        newData = []
        for elem in data:
            if elem == Slip.SLIP_BYTE_END:
                newData.append(Slip.SLIP_BYTE_ESC)
                newData.append(Slip.SLIP_BYTE_ESC_END)
            elif elem == Slip.SLIP_BYTE_ESC:
                newData.append(Slip.SLIP_BYTE_ESC)
                newData.append(Slip.SLIP_BYTE_ESC_ESC)
            else:
                newData.append(elem)
        newData.append(Slip.SLIP_BYTE_END)
        return newData

    @staticmethod
    def decode_add_byte(c, decoded_data, current_state):
        finished = False
        if current_state == Slip.SLIP_STATE_DECODING:
            if c == Slip.SLIP_BYTE_END:
                finished = True
            elif c == Slip.SLIP_BYTE_ESC:
                current_state = Slip.SLIP_STATE_ESC_RECEIVED
            else:
                decoded_data.append(c)
        elif current_state == Slip.SLIP_STATE_ESC_RECEIVED:
            if c == Slip.SLIP_BYTE_ESC_END:
                decoded_data.append(Slip.SLIP_BYTE_END)
                current_state = Slip.SLIP_STATE_DECODING
            elif c == Slip.SLIP_BYTE_ESC_ESC:
                decoded_data.append(Slip.SLIP_BYTE_ESC)
                current_state = Slip.SLIP_STATE_DECODING
            else:
                current_state = Slip.SLIP_STATE_CLEARING_INVALID_PACKET
        elif current_state == Slip.SLIP_STATE_CLEARING_INVALID_PACKET:
            if c == Slip.SLIP_BYTE_END:
                current_state = Slip.SLIP_STATE_DECODING
                decoded_data = []

        return (finished, current_state, decoded_data)

class DFUAdapter:
    def __init__(self, serial_port):
        self.serial_port = serial_port

    def send_message(self, data):
        packet = Slip.encode(data)
        logger.log(TRANSPORT_LOGGING_LEVEL, 'SLIP: --> ' + str(data))
        try:
            self.serial_port.write(packet)
        except SerialException as e:
            raise NordicSemiException('Writing to serial port failed: ' + str(e) + '. '
                                      'If MSD is enabled on the target device, try to disable it ref. '
                                      'https://wiki.segger.com/index.php?title=J-Link-OB_SAM3U')

    def get_message(self):
        current_state = Slip.SLIP_STATE_DECODING
        finished = False
        decoded_data = []

        while finished == False:
            byte = self.serial_port.read(1)
            if byte:
                (byte) = struct.unpack('B', byte)[0]
                (finished, current_state, decoded_data) \
                   = Slip.decode_add_byte(byte, decoded_data, current_state)
            else:
                current_state = Slip.SLIP_STATE_CLEARING_INVALID_PACKET
                return None

        logger.log(TRANSPORT_LOGGING_LEVEL, 'SLIP: <-- ' + str(decoded_data))

        return decoded_data

class DfuTransportSerial(DfuTransport):

    DEFAULT_BAUD_RATE = 115200
    DEFAULT_FLOW_CONTROL = True
    DEFAULT_TIMEOUT = 30.0  # Timeout time for board response
    DEFAULT_SERIAL_PORT_TIMEOUT = 1.0  # Timeout time on serial port read
    DEFAULT_PRN                 = 0
    DEFAULT_DO_PING = True

    OP_CODE = {
        'CreateObject'          : 0x01,
        'SetPRN'                : 0x02,
        'CalcChecSum'           : 0x03,
        'Execute'               : 0x04,
        'ReadError'             : 0x05,
        'ReadObject'            : 0x06,
        'GetSerialMTU'          : 0x07,
        'WriteObject'           : 0x08,
        'Ping'                  : 0x09,
        'Response'              : 0x60,
    }

    def __init__(self,
                 com_port,
                 baud_rate=DEFAULT_BAUD_RATE,
                 flow_control=DEFAULT_FLOW_CONTROL,
                 timeout=DEFAULT_TIMEOUT,
                 prn=DEFAULT_PRN,
                 do_ping=DEFAULT_DO_PING):

        super().__init__()
        self.com_port = com_port
        self.baud_rate = baud_rate
        self.flow_control = 1 if flow_control else 0
        self.timeout = timeout
        self.prn         = prn
        self.serial_port = None
        self.dfu_adapter = None
        self.ping_id     = 0
        self.do_ping     = do_ping

        self.mtu         = 0

        """:type: serial.Serial """


    def open(self):
        super().open()
        try:
            self.__ensure_bootloader()
            self.serial_port = Serial(port=self.com_port,
                baudrate=self.baud_rate, rtscts=self.flow_control, timeout=self.DEFAULT_SERIAL_PORT_TIMEOUT)
            self.dfu_adapter = DFUAdapter(self.serial_port)
        except OSError as e:
            raise NordicSemiException("Serial port could not be opened on {0}"
              ". Reason: {1}".format(self.com_port, e.strerror))

        if self.do_ping:
            ping_success = False
            start = datetime.now()
            while (datetime.now() - start < timedelta(seconds=self.timeout)
                    and ping_success == False):
                if self.__ping() == True:
                    ping_success = True

            if ping_success == False:
                raise NordicSemiException("No ping response after opening COM port")

        self.__set_prn()
        self.__get_mtu()

    def close(self):
        super().close()
        self.serial_port.close()

    def send_init_packet(self, init_packet):
        def try_to_recover():
            if response['offset'] == 0 or response['offset'] > len(init_packet):
                # There is no init packet or present init packet is too long.
                return False

            expected_crc = (binascii.crc32(init_packet[:response['offset']]) & 0xFFFFFFFF)

            if expected_crc != response['crc']:
                # Present init packet is invalid.
                return False

            if len(init_packet) > response['offset']:
                # Send missing part.
                try:
                    self.__stream_data(data     = init_packet[response['offset']:],
                                       crc      = expected_crc,
                                       offset   = response['offset'])
                except ValidationException:
                    return False

            self.__execute()
            return True

        response = self.__select_command()
        assert len(init_packet) <= response['max_size'], 'Init command is too long'

        if try_to_recover():
            return

        try:
            self.__create_command(len(init_packet))
            self.__stream_data(data=init_packet)
            self.__execute()
        except ValidationException:
            raise NordicSemiException("Failed to send init packet")

    def send_firmware(self, firmware):
        def try_to_recover():
            if response['offset'] == 0:
                # Nothing to recover
                return

            expected_crc = binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
            remainder    = response['offset'] % response['max_size']

            if expected_crc != response['crc']:
                # Invalid CRC. Remove corrupted data.
                response['offset'] -= remainder if remainder != 0 else response['max_size']
                response['crc']     = \
                        binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
                return

            if (remainder != 0) and (response['offset'] != len(firmware)):
                # Send rest of the page.
                try:
                    to_send             = firmware[response['offset'] : response['offset']
                                                + response['max_size'] - remainder]
                    response['crc']     = self.__stream_data(data   = to_send,
                                                             crc    = response['crc'],
                                                             offset = response['offset'])
                    response['offset'] += len(to_send)
                except ValidationException:
                    # Remove corrupted data.
                    response['offset'] -= remainder
                    response['crc']     = \
                        binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
                    return

            self.__execute()
            self._send_event(event_type=DfuEvent.PROGRESS_EVENT, progress=response['offset'])

        response = self.__select_data()
        try_to_recover()
        for i in range(response['offset'], len(firmware), response['max_size']):
            data = firmware[i:i+response['max_size']]
            try:
                self.__create_data(len(data))
                response['crc'] = self.__stream_data(data=data, crc=response['crc'], offset=i)
                self.__execute()
            except ValidationException:
                raise NordicSemiException("Failed to send firmware")

            self._send_event(event_type=DfuEvent.PROGRESS_EVENT, progress=len(data))

    def __ensure_bootloader(self):
        lister = DeviceLister()

        device = None
        start = datetime.now()
        while not device and datetime.now() - start < timedelta(seconds=self.timeout):
            time.sleep(0.5)
            device = lister.get_device(com=self.com_port)

        if device:
            device_serial_number = device.serial_number

            if not self.__is_device_in_bootloader_mode(device):
                retry_count = 10
                wait_time_ms = 500

                trigger = DFUTrigger()
                try:
                    trigger.enter_bootloader_mode(device)
                    logger.info("Serial: DFU bootloader was triggered")
                except NordicSemiException as err:
                    logger.error(err)


                for checks in range(retry_count):
                    logger.info("Serial: Waiting {} ms for device to enter bootloader {}/{} time"\
                    .format(500, checks + 1, retry_count))

                    time.sleep(wait_time_ms / 1000.0)

                    device = lister.get_device(serial_number=device_serial_number)
                    if self.__is_device_in_bootloader_mode(device):
                        self.com_port = device.get_first_available_com_port()
                        break

                trigger.clean()
            if not self.__is_device_in_bootloader_mode(device):
                logger.info("Serial: Device is either not in bootloader mode, or using an unsupported bootloader.")

    def __is_device_in_bootloader_mode(self, device):
        if not device:
            return False

        #  Return true if nrf bootloader or Jlink interface detected.
        return ((device.vendor_id.lower() == '1915' and device.product_id.lower() == '521f') # nRF52 SDFU USB
             or (device.vendor_id.lower() == '1366' and device.product_id.lower() == '0105') # JLink CDC UART Port
             or (device.vendor_id.lower() == '1366' and device.product_id.lower() == '1015'))# JLink CDC UART Port (MSD)

    def __set_prn(self):
        logger.debug("Serial: Set Packet Receipt Notification {}".format(self.prn))
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE['SetPRN']]
            + list(struct.pack('<H', self.prn)))
        self.__get_response(DfuTransportSerial.OP_CODE['SetPRN'])

    def __get_mtu(self):
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE['GetSerialMTU']])
        response = self.__get_response(DfuTransportSerial.OP_CODE['GetSerialMTU'])

        self.mtu = struct.unpack('<H', bytearray(response))[0]

    def __ping(self):
        self.ping_id = (self.ping_id + 1) % 256

        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE['Ping'], self.ping_id])
        resp = self.dfu_adapter.get_message() # Receive raw response to check return code

        if not resp:
            logger.debug('Serial: No ping response')
            return False

        if resp[0] != DfuTransportSerial.OP_CODE['Response']:
            logger.debug('Serial: No Response: 0x{:02X}'.format(resp[0]))
            return False

        if resp[1] != DfuTransportSerial.OP_CODE['Ping']:
            logger.debug('Serial: Unexpected Executed OP_CODE.\n' \
                + 'Expected: 0x{:02X} Received: 0x{:02X}'.format(DfuTransportSerial.OP_CODE['Ping'], resp[1]))
            return False

        if resp[2] != DfuTransport.RES_CODE['Success']:
            # Returning an error code is seen as good enough. The bootloader is up and running
            return True
        else:
            if struct.unpack('B', bytearray(resp[3:]))[0] == self.ping_id:
                return True
            else:
                return False

    def __create_command(self, size):
        self.__create_object(0x01, size)

    def __create_data(self, size):
        self.__create_object(0x02, size)

    def __create_object(self, object_type, size):
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE['CreateObject'], object_type]\
                                            + list(struct.pack('<L', size)))
        self.__get_response(DfuTransportSerial.OP_CODE['CreateObject'])

    def __calculate_checksum(self):
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE['CalcChecSum']])
        response = self.__get_response(DfuTransportSerial.OP_CODE['CalcChecSum'])

        if response is None:
            raise NordicSemiException('Did not receive checksum response from DFU target. '
                                      'If MSD is enabled on the target device, try to disable it ref. '
                                      'https://wiki.segger.com/index.php?title=J-Link-OB_SAM3U')

        (offset, crc) = struct.unpack('<II', bytearray(response))
        return {'offset': offset, 'crc': crc}

    def __execute(self):
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE['Execute']])
        self.__get_response(DfuTransportSerial.OP_CODE['Execute'])

    def __select_command(self):
        return self.__select_object(0x01)

    def __select_data(self):
        return self.__select_object(0x02)

    def __select_object(self, object_type):
        logger.debug("Serial: Selecting Object: type:{}".format(object_type))
        self.dfu_adapter.send_message([DfuTransportSerial.OP_CODE['ReadObject'], object_type])

        response = self.__get_response(DfuTransportSerial.OP_CODE['ReadObject'])
        (max_size, offset, crc)= struct.unpack('<III', bytearray(response))

        logger.debug("Serial: Object selected: " +
            " max_size:{} offset:{} crc:{}".format(max_size, offset, crc))
        return {'max_size': max_size, 'offset': offset, 'crc': crc}

    def __get_checksum_response(self):
        resp = self.__get_response(DfuTransportSerial.OP_CODE['CalcChecSum'])

        (offset, crc) = struct.unpack('<II', bytearray(resp))
        return {'offset': offset, 'crc': crc}

    def __stream_data(self, data, crc=0, offset=0):
        logger.debug("Serial: Streaming Data: " +
            "len:{0} offset:{1} crc:0x{2:08X}".format(len(data), offset, crc))
        def validate_crc():
            if (crc != response['crc']):
                raise ValidationException('Failed CRC validation.\n'\
                                + 'Expected: {} Received: {}.'.format(crc, response['crc']))
            if (offset != response['offset']):
                raise ValidationException('Failed offset validation.\n'\
                                + 'Expected: {} Received: {}.'.format(offset, response['offset']))

        current_pnr     = 0

        for i in range(0, len(data), (self.mtu-1)//2 - 1):
            # append the write data opcode to the front
            # here the maximum data size is self.mtu/2,
            # due to the slip encoding which at maximum doubles the size
            to_transmit = data[i:i + (self.mtu-1)//2 - 1 ]
            to_transmit = struct.pack('B',DfuTransportSerial.OP_CODE['WriteObject']) + to_transmit

            self.dfu_adapter.send_message(list(to_transmit))
            crc     = binascii.crc32(to_transmit[1:], crc) & 0xFFFFFFFF
            offset += len(to_transmit) - 1
            current_pnr    += 1
            if self.prn == current_pnr:
                current_pnr = 0
                response    = self.__get_checksum_response()
                validate_crc()
        response = self.__calculate_checksum()
        validate_crc()
        return crc

    def __get_response(self, operation):
        def get_dict_key(dictionary, value):
            return next((key for key, val in list(dictionary.items()) if val == value), None)

        resp = self.dfu_adapter.get_message()

        if not resp:
            return None

        if resp[0] != DfuTransportSerial.OP_CODE['Response']:
            raise NordicSemiException('No Response: 0x{:02X}'.format(resp[0]))

        if resp[1] != operation:
            raise NordicSemiException('Unexpected Executed OP_CODE.\n' \
                             + 'Expected: 0x{:02X} Received: 0x{:02X}'.format(operation, resp[1]))

        if resp[2] == DfuTransport.RES_CODE['Success']:
            return resp[3:]

        elif resp[2] == DfuTransport.RES_CODE['ExtendedError']:
            try:
                data = DfuTransport.EXT_ERROR_CODE[resp[3]]
            except IndexError:
                data = "Unsupported extended error type {}".format(resp[3])
            raise NordicSemiException('Extended Error 0x{:02X}: {}'.format(resp[3], data))
        else:
            raise NordicSemiException('Response Code {}'.format(
                get_dict_key(DfuTransport.RES_CODE, resp[2])))



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import time
import logging
import ipaddress
import struct

import io
import spinel.common
import spinel.ipv6

from spinel.stream import StreamOpen
from spinel.codec import WpanApi
from spinel.const import SPINEL
import collections
import spinel.util as util

logger = logging.getLogger(__name__)

class NCPTransport:
    '''A CoAP Toolkit compatible transport'''
    CFG_KEY_CHANNEL = 'channel'
    CFG_KEY_PANID = 'panid'
    CFG_KEY_RESET = 'reset'
    CFG_KEY_MASTERKEY = 'masterkey'

    def __init__(self, port, stream_descriptor, config = None):
        self._port = port
        self._stream_descriptor = stream_descriptor.split(":")
        self._config = config if config is not None else self.get_default_config()
        self._attached = False

        self._udp6_parser = spinel.ipv6.IPv6PacketFactory(
                            ulpf = {
                                17: spinel.ipv6.UDPDatagramFactory(
                                    udp_header_factory = spinel.ipv6.UDPHeaderFactory(),
                                    dst_port_factories = {
                                        port: spinel.ipv6.UDPBytesPayloadFactory()
                                        }
                                    ),
                            })

        self._receivers = []

    @staticmethod
    def _propid_to_str(propid):
        for name, value in SPINEL.__dict__.items():
            if (name.startswith('PROP_') and value == propid):
                return name

    def _set_property(self, *args):
        for propid, value, py_format in args:
            logger.debug("Setting property %s to %s", self.__class__._propid_to_str(propid), str(value))
            result = self._wpan.prop_set_value(propid, value, py_format)
            if (result is None):
                raise Exception("Failed to set property {}".format(self.__class__._propid_to_str(propid)))
            else:
                logger.debug("Done")

    def _get_property(self, *args):
        return self._wpan.prop_get_value(SPINEL.PROP_LAST_STATUS)

    def _attach_to_network(self):
        props = [
            (SPINEL.PROP_IPv6_ICMP_PING_OFFLOAD, 1, 'B'),
            (SPINEL.PROP_THREAD_RLOC16_DEBUG_PASSTHRU, 1, 'B'),
            (SPINEL.PROP_PHY_CHAN, self._config[self.CFG_KEY_CHANNEL], 'H'),
            (SPINEL.PROP_MAC_15_4_PANID, self._config[self.CFG_KEY_PANID], 'H'),
            (SPINEL.PROP_NET_MASTER_KEY, self._config[self.CFG_KEY_MASTERKEY], '16s'),
            (SPINEL.PROP_NET_IF_UP, 1, 'B'),
            (SPINEL.PROP_NET_STACK_UP, 1, 'B'),
        ]
        self._set_property(*props)

        while True:
            role = self._wpan.prop_get_value(SPINEL.PROP_NET_ROLE)
            if (role != 0):
                self._attached = True
                return True
            time.sleep(1)

        return False

    def _wpan_receive(self, prop, value, tid):
        consumed = False
        if prop == SPINEL.PROP_STREAM_NET:
            consumed = True
            try:
                pkt = self._udp6_parser.parse(io.BytesIO(value),
                                              spinel.common.MessageInfo())
                endpoint = collections.namedtuple('endpoint', 'addr port')
                payload = pkt.upper_layer_protocol.payload.to_bytes()
                src = endpoint(pkt.ipv6_header.source_address,
                               pkt.upper_layer_protocol.header.src_port)
                dst = endpoint(pkt.ipv6_header.destination_address,
                               pkt.upper_layer_protocol.header.dst_port)

                for receiver in self._receivers:
                    receiver.receive(payload, src, dst)

            except RuntimeError:
                pass
            except Exception as e:
                logging.exception(e)
        else:
            logger.warning("Unexpected property received (PROP_ID: {})".format(prop))

        return consumed

    def _build_udp_datagram(self, saddr, sport, daddr, dport, payload):
        return spinel.ipv6.IPv6Packet(spinel.ipv6.IPv6Header(saddr, daddr),
                                      spinel.ipv6.UDPDatagram(
                                          spinel.ipv6.UDPHeader(sport, dport),
                                          spinel.ipv6.UDPBytesPayload(payload)))

    @classmethod
    def get_default_config(cls):
        return {cls.CFG_KEY_CHANNEL:     11,
                cls.CFG_KEY_PANID:       0xabcd,
                cls.CFG_KEY_RESET:       True,
                cls.CFG_KEY_MASTERKEY:   util.hex_to_bytes("00112233445566778899aabbccddeeff")}

    def add_ip_address(self, ipaddr):
        valid = 1
        preferred = 1
        flags = 0
        prefix_len = 64

        prefix = ipaddress.IPv6Interface(str(ipaddr))
        arr = prefix.ip.packed
        arr += self._wpan.encode_fields('CLLC',
                                           prefix_len,
                                           valid,
                                           preferred,
                                           flags)

        self._wpan.prop_insert_value(SPINEL.PROP_IPV6_ADDRESS_TABLE, arr, str(len(arr)) + 's')
        logger.debug("Added")

    def print_addresses(self):
        logger.info("NCP Thread IPv6 addresses:")
        for addr in self._wpan.get_ipaddrs():
            logger.info(str(addr))

    def send(self, payload, dest):
        if (dest.addr.is_multicast):
            rloc16 = self._wpan.prop_get_value(SPINEL.PROP_THREAD_RLOC16)

            # Create an IPv6 Thread RLOC address from mesh-local prefix and RLOC16 MAC address.
            src_addr = ipaddress.ip_address(self._ml_prefix + b'\x00\x00\x00\xff\xfe\x00' + struct.pack('>H', rloc16))

        else:
            src_addr = self._ml_eid

        logger.debug("Sending datagram {} {} {} {}".format(src_addr,
                                                           self._port,
                                                           dest.addr,
                                                           dest.port))
        try:
            datagram = self._build_udp_datagram(src_addr,
                                                self._port,
                                                dest.addr,
                                                dest.port,
                                                payload)
        except Exception as e:
            logging.exception(e)

        self._wpan.ip_send(datagram.to_bytes())

    def register_receiver(self, callback):
        '''Registers a receiver, that will get all the data received from the transport.
           The callback function shall be in format:
           receive_callback(src_addr, src_port, dest_port, dest_addr, payload)'''
        self._receivers.append(callback)

    def remove_receiver(self, callback):
        '''Remove a receiver callback'''
        self._receivers.remove(callback)

    def open(self):
        '''Opens transport for communication.'''
        self._stream = StreamOpen(self._stream_descriptor[0], self._stream_descriptor[1], False)
        # FIXME: remove node id from constructor after WpanAPI is refactored
        self._wpan = WpanApi(self._stream, 666)
        self._wpan.queue_register(SPINEL.HEADER_DEFAULT)
        self._wpan.queue_register(SPINEL.HEADER_ASYNC)
        self._wpan.callback_register(SPINEL.PROP_STREAM_NET, self._wpan_receive)

        if (self._config[NCPTransport.CFG_KEY_RESET]) and not self._wpan.cmd_reset():
            raise Exception('Failed to reset NCP. Please flash connectivity firmware.')

        logger.info('Attaching to the network')
        if (not self._attach_to_network()):
            logger.error("Failed to attach to the network")
            raise Exception('Unable to attach')

        self._ml_eid = ipaddress.ip_address(self._wpan.prop_get_value(SPINEL.PROP_IPV6_ML_ADDR))
        self._ml_prefix = self._wpan.prop_get_value(SPINEL.PROP_IPV6_ML_PREFIX)

        logger.info("Done")

        self.print_addresses()

    def close(self):
        '''Closes transport for communication.'''
        self._wpan.cmd_reset()
        self._stream.close()



"""
MIT License

Copyright (c) 2016 gwangyi

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import enum
from .structures import DevicePropertyKey


# noinspection SpellCheckingInspection
class DiOpenDeviceInfo(enum.IntEnum):
    """DIOD_xxx constants
    """
    InheritClassDrvs = 2
    CancelRemove = 4


# noinspection SpellCheckingInspection
class DiGetClassDevsFlags(enum.IntEnum):
    """DIGCF_xxx constants
    """
    Default = 0x00000001
    Present = 0x00000002,
    AllClasses = 0x00000004,
    Profile = 0x00000008,
    DeviceInterface = 0x00000010,


# noinspection SpellCheckingInspection
class DevicePropertyKeys:
    """DEVPKEY_xxx constants"""
    NAME = DevicePropertyKey('{b725f130-47ef-101a-a5f1-02608c9eebac}', 10, 'DEVPKEY_NAME')
    Numa_Proximity_Domain = DevicePropertyKey('{540b947e-8b40-45bc-a8a2-6a0b894cbda2}', 1,
                                              'DEVPKEY_Numa_Proximity_Domain')

    # noinspection SpellCheckingInspection
    class Device:
        """DEVPKEY_Device_xxx constants"""
        ContainerId = DevicePropertyKey('{8c7ed206-3f8a-4827-b3ab-ae9e1faefc6c}', 2,
                                        'DEVPKEY_Device_ContainerId')
        DeviceAddress = DevicePropertyKey('{a45c254e-df1c-4efd-8020-67d146a850e0}', 30,
                                        'DEVPKEY_Device_Address')

# noinspection SpellCheckingInspection
DIGCF_DEFAULT = DiGetClassDevsFlags.Default
# noinspection SpellCheckingInspection
DIGCF_PRESENT = DiGetClassDevsFlags.Present
# noinspection SpellCheckingInspection
DIGCF_ALLCLASSES = DiGetClassDevsFlags.AllClasses
# noinspection SpellCheckingInspection
DIGCF_PROFILE = DiGetClassDevsFlags.Profile
# noinspection SpellCheckingInspection
DIGCF_DEVICEINTERFACE = DiGetClassDevsFlags.DeviceInterface

# noinspection SpellCheckingInspection
DIOD_INHERIT_CLASSDRVS = DiOpenDeviceInfo.InheritClassDrvs
# noinspection SpellCheckingInspection
DIOD_CANCEL_REMOVE = DiOpenDeviceInfo.CancelRemove

# noinspection SpellCheckingInspection
DEVPKEY = DevicePropertyKeys
DEVPKEY_Device_ContainerId = DevicePropertyKeys.Device.ContainerId



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import binascii
import os
import shutil
import tempfile
import unittest

from nordicsemi.dfu.signing import Signing


class TestSigning(unittest.TestCase):
    def setUp(self):
        script_abspath = os.path.abspath(__file__)
        script_dirname = os.path.dirname(script_abspath)
        os.chdir(script_dirname)

    def test_gen_key(self):
        self.work_directory = tempfile.mkdtemp(prefix="nrf_signing_tests_")

        key_file_name = 'key.pem'
        key_file_path = os.path.join(self.work_directory, key_file_name)

        signing = Signing()
        signing.gen_key(key_file_path)

        self.assertTrue(os.path.exists(key_file_path))

        shutil.rmtree(self.work_directory, ignore_errors=True)

    def test_load_key(self):
        key_file_name = 'key.pem'

        signing = Signing()
        signing.load_key(key_file_name)

        self.assertEqual(64, len(binascii.hexlify(signing.sk.to_string())))

    def test_get_vk(self):
        key_file_name = 'key.pem'

        signing = Signing()
        signing.load_key(key_file_name)

        vk_str = signing.get_vk('hex', False)
        vk_hex = signing.get_vk_hex()
        self.assertEqual(vk_hex, vk_str)

        vk_str = signing.get_vk('code', False)
        vk_code = signing.get_vk_code(False)
        self.assertEqual(vk_code, vk_str)

        vk_str = signing.get_vk('pem', False)
        vk_pem = signing.get_vk_pem()
        self.assertEqual(vk_pem, vk_str)

    def test_get_vk_hex(self):
        key_file_name = 'key.pem'
        expected_vk_hex = "Public (verification) key pk:\n60f417aabb6bb5b9058aec0570b83fedab1782d62072ae7d691f98dbeda28d654c1d98"\
                          "6cadcd593ad8901084900c1bbdcc4fff62b612b604c22672adcdae9b90"

        signing = Signing()
        signing.load_key(key_file_name)

        vk_hex = signing.get_vk_hex()

        self.assertEqual(expected_vk_hex, vk_hex)

    def test_get_sk_hex(self):
        key_file_name = 'key.pem'
        expected_vk_hex = "Private (signing) key sk:\n9d37cc501be62612dd44d150a1c6ad69fbf08e9f905e5e860b89ff9e1050963d"

        signing = Signing()
        signing.load_key(key_file_name)

        sk_hex = signing.get_sk_hex()

        self.assertEqual(expected_vk_hex, sk_hex)

    def test_get_vk_pem(self):
        key_file_name = 'key.pem'
        expected_vk_pem = "-----BEGIN PUBLIC KEY-----\n" \
                          "MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEZY2i7duYH2l9rnIg1oIXq+0/uHAF\n" \
                          "7IoFubVru6oX9GCQm67NrXImwgS2ErZi/0/MvRsMkIQQkNg6Wc2tbJgdTA==\n" \
                          "-----END PUBLIC KEY-----\n"

        signing = Signing()
        signing.load_key(key_file_name)

        vk_pem = signing.get_vk_pem()

        self.assertEqual(expected_vk_pem, vk_pem)



#
# Copyright (c) 2016-2018 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# Python imports
import binascii
from datetime import datetime
import logging
import queue
import struct
import sys


# Python 3rd party imports
try:
    import antlib
    import antlib.antdefines as antdefines
    import antlib.antmessage as antmessage
except ImportError as e:
    print(e)
    raise Exception("Try running 'pip install antlib'.")

# Nordic Semiconductor imports
from nordicsemi.dfu.dfu_transport   import DfuTransport, DfuEvent, TRANSPORT_LOGGING_LEVEL
from pc_ble_driver_py.exceptions    import NordicSemiException


def platform_supported():
    """
    A platform check should be performed before using this module.
    Only a python 32bit interpreter on windows is supported. The 'antlib'
    package depends on precompiled windows 32bit libraries.

    :return: True if platform is windows 32bit or False otherwise.
    :rtype: bool
    """
    can_run = False

    if sys.platform in ["win32", "win64"]:
        # Check if Python is 32 bit.
        if struct.calcsize("P") * 8 == 32:
            can_run = True

    if not can_run:
        print("The ant dfu command is only available on Windows with a 32bit Python \n"
              "which can be downloaded here: 'https://www.python.org/downloads/windows/'.")

    return can_run


class ValidationException(NordicSemiException):
    """"
    Exception used when validation failed
    """
    pass

logger = logging.getLogger(__name__)


class AntParams:
    # 2466 MHz
    DEF_RF_FREQ = 66
    # 16 Hz
    DEF_CHANNEL_PERIOD = 2048
    # Wildcard specific device.
    DEF_DEVICE_NUM = 0
    DEF_DEVICE_TYPE = 10
    DEF_TRANS_TYPE = 0
    DEF_NETWORK_KEY = None

    def __init__(self):
        self.rf_freq = self.DEF_RF_FREQ
        self.channel_period = self.DEF_CHANNEL_PERIOD
        self.device_num = self.DEF_DEVICE_NUM
        self.device_type = self.DEF_DEVICE_TYPE
        self.trans_type = self.DEF_TRANS_TYPE
        self.network_key = self.DEF_NETWORK_KEY


class DfuAdapter:
    ANT_RSP_TIMEOUT = 100
    ANT_DFU_CHAN = 0
    ANT_NET_KEY_IDX = 0

    DATA_MESGS = (
        antmessage.MESG_BROADCAST_DATA_ID,
        antmessage.MESG_ACKNOWLEDGED_DATA_ID,
        antmessage.MESG_BURST_DATA_ID,
        antmessage.MESG_ADV_BURST_DATA_ID)

    def __init__(self, ant_dev, timeout, search_timeout, ant_config):
        self.ant_dev = ant_dev
        self.timeout = timeout
        self.search_timeout = search_timeout
        self.ant_config = ant_config
        self.connected = False
        self.tx_result = None
        self.beacon_rx = False
        self.tx_seq = None
        self.rx_seq = None
        self.rx_data = None
        self.resp_queue = queue.Queue()

    def open(self):
        # TODO: use constant from antlib when it exists.
        # Set up advanced burst with optional frequency hopping. This can help
        # Increase the throughput by about 3x.
        self.ant_dev.configure_advanced_burst(True, 3, 0, 0x01,
            response_time_msec=self.ANT_RSP_TIMEOUT)

        if self.ant_config.network_key is not None:
            self.ant_dev.set_network_key(self.ANT_NET_KEY_IDX,
                self.ant_config.network_key)

        self.ant_dev.assign_channel(self.ANT_DFU_CHAN, antdefines.CHANNEL_TYPE_SLAVE,
            self.ANT_NET_KEY_IDX, response_time_msec=self.ANT_RSP_TIMEOUT)

        # The params here are fairly arbitrary, but must match what the device
        # uses. These match the defaults in the example projects.
        self.ant_dev.set_channel_freq(self.ANT_DFU_CHAN,
            self.ant_config.rf_freq, response_time_msec=self.ANT_RSP_TIMEOUT)

        self.ant_dev.set_channel_period(self.ANT_DFU_CHAN,
            self.ant_config.channel_period,
            response_time_msec=self.ANT_RSP_TIMEOUT)

        self.ant_dev.set_channel_id(self.ANT_DFU_CHAN,
            self.ant_config.device_num, self.ant_config.device_type,
            self.ant_config.trans_type, response_time_msec=self.ANT_RSP_TIMEOUT)

        # Disable high priority search. It doesn't matter much which search
        # is used as there are no other open channels, but configuring only
        # one makes sure the timeout arrives when expected.
        self.ant_dev.set_channel_search_timeout(self.ANT_DFU_CHAN, 0,
            response_time_msec=self.ANT_RSP_TIMEOUT)

        self.ant_dev.set_low_priority_search_timeout(
            self.ANT_DFU_CHAN,
            int(self.search_timeout / 2.5),
            response_time_msec=self.ANT_RSP_TIMEOUT)

        self.ant_dev.open_channel(self.ANT_DFU_CHAN,
            response_time_msec=self.ANT_RSP_TIMEOUT)

        self.__wait_for_condition(
            lambda: self.connected, self.search_timeout + 1.0)

    def close(self):
        self.ant_dev.reset_system(response_time_msec=self.ANT_RSP_TIMEOUT)
        self.ant_dev.ant_close()
        self.ant_dev = None

    def send_message(self, req):
        logger.log(TRANSPORT_LOGGING_LEVEL, "ANT: --> {}".format(req))

        self.tx_seq = (self.tx_seq + 1) & 0xFF
        data = list(struct.pack('<HB', len(req) + 3, self.tx_seq)) + req

        self.tx_result = None

        while self.tx_result != True:
            self.tx_result = None

            self.ant_dev.send_burst(self.ANT_DFU_CHAN, data)
            self.__wait_for_condition(lambda: self.tx_result is not None)

            # Wait for a beacon, needed in tx fail case to allow for flush of
            # any sequence number errors that could interrupt the burst.
            self.beacon_rx = False
            self.__wait_for_condition(lambda: self.beacon_rx)

    def get_message(self):
        self.__wait_for_condition(lambda: not self.resp_queue.empty())
        mesg = self.resp_queue.get_nowait()

        logger.log(TRANSPORT_LOGGING_LEVEL, "ANT: <-- {}".format(mesg))
        return mesg

    def __wait_for_condition(self, cond, timeout=None):
        if timeout is None:
            timeout = self.timeout

        start = datetime.now()
        while not cond():
            self.__process_mesg(timeout -
                (datetime.now() - start).total_seconds())

    def __process_mesg(self, timeout):
        mesg = self.__get_ant_mesg(timeout)

        if (mesg.msgid in self.DATA_MESGS):
            self.__process_data_mesg(mesg)
        elif (mesg.msgid == antmessage.MESG_RESPONSE_EVENT_ID):
            if (mesg.data[1] == antmessage.MESG_EVENT_ID):
                self.__process_evt(mesg.data[2])

    def __process_data_mesg(self, mesg):
        if (mesg.msgid == antmessage.MESG_BROADCAST_DATA_ID):
            self.connected = True
            self.beacon_rx = True
            # Broadcast data should always contain the current sequence numbers.
            if (self.tx_seq is None):
                self.tx_seq = mesg.data[1]
            if (self.rx_seq is None):
                self.rx_seq = mesg.data[2]
            return

        if not self.connected:
            # Ignore non-broadcast data until connection is established.
            return

        data = mesg.data[1:1+antdefines.ANT_STANDARD_DATA_PAYLOAD_SIZE]
        is_first = False
        is_last = False

        if (mesg.msgid == antmessage.MESG_ACKNOWLEDGED_DATA_ID):
            is_first = is_last = True
        else:
            if mesg.sequence_number == antdefines.SEQUENCE_FIRST_MESSAGE:
                is_first = True
            if (mesg.sequence_number & antdefines.SEQUENCE_LAST_MESSAGE) != 0:
                is_last = True
            if (mesg.msgid == antmessage.MESG_ADV_BURST_DATA_ID):
                data = mesg.data[1:]

        if is_first:
            self.rx_data = data
        else:
            self.rx_data += data

        if is_last:
            self.__process_resp()
            self.rx_data = None

    def __process_resp(self):
        (size, seq) = struct.unpack('<HB', bytearray(self.rx_data[:3]))

        if seq == self.rx_seq:
            logger.debug("Duplicate response received")
            return

        self.rx_seq = seq
        self.resp_queue.put(self.rx_data[3:size])

    def __process_evt(self, evt):
        if (evt == antdefines.EVENT_CHANNEL_CLOSED):
            raise NordicSemiException("Device connection lost")
        elif (evt == antdefines.EVENT_TRANSFER_TX_COMPLETED):
            self.tx_result = True
        elif (evt == antdefines.EVENT_TRANSFER_TX_FAILED):
            self.tx_result = False
        elif (evt == antdefines.EVENT_TRANSFER_RX_FAILED):
            self.rx_data = None

    def __get_ant_mesg(self, timeout):
        mesg = None
        if self.ant_dev.wait_for_message(int(timeout * 1000)):
            mesg = self.ant_dev.get_message()

        if not mesg:
            raise NordicSemiException('No message received from device')

        return mesg


class DfuTransportAnt(DfuTransport):
    ANT_RST_TIMEOUT_MS          = 500
    DEFAULT_PORT                = 0
    DEFAULT_CMD_TIMEOUT         = 5.0   # Timeout on waiting for a response.
    DEFAULT_SEARCH_TIMEOUT      = 10.0  # Timeout when searching for the device.
    DEFAULT_PRN                 = 0
    DEFAULT_DO_PING             = True
    DEFAULT_DO_DEBUG            = False

    OP_CODE = {
        'CreateObject'          : 0x01,
        'SetPRN'                : 0x02,
        'CalcChecSum'           : 0x03,
        'Execute'               : 0x04,
        'ReadError'             : 0x05,
        'ReadObject'            : 0x06,
        'GetSerialMTU'          : 0x07,
        'WriteObject'           : 0x08,
        'Ping'                  : 0x09,
        'Response'              : 0x60,
    }

    def __init__(self,
                 ant_config=None,
                 port=DEFAULT_PORT,
                 timeout=DEFAULT_CMD_TIMEOUT,
                 search_timeout=DEFAULT_SEARCH_TIMEOUT,
                 prn=DEFAULT_PRN,
                 debug=DEFAULT_DO_DEBUG):

        super().__init__()
        if ant_config is None:
            ant_config = AntParams()
        self.ant_config     = ant_config
        self.port           = port
        self.timeout        = timeout
        self.search_timeout = search_timeout
        self.prn            = prn
        self.ping_id        = 0
        self.dfu_adapter    = None
        self.mtu            = 0
        self.debug          = debug


    def open(self):
        super().open()
        ant_dev = None
        try:
            ant_dev = antlib.ANTDevice(self.port, 57600,
                antlib.antdevice.ANTDevice.USB_PORT_TYPE,
                antlib.antdevice.ANTDevice.FRAMER_TYPE_BASIC)
            ant_dev.reset_system(response_time_msec=self.ANT_RST_TIMEOUT_MS)
            # There might be a back log of messages, clear them out.
            while ant_dev.get_message():
                pass
            if (self.debug):
                ant_dev.enable_debug_logging()
        except Exception as e:
            raise NordicSemiException(
                "Could not open {0}. Reason: {1}".format(ant_dev, e.message))

        self.dfu_adapter = DfuAdapter(
            ant_dev, self.timeout, self.search_timeout, self.ant_config)
        self.dfu_adapter.open()

        if not self.__ping():
            raise NordicSemiException("No ping response from device.")

        self.__set_prn()
        self.__get_mtu()

    def close(self):
        super().close()
        self.dfu_adapter.close()

    def send_init_packet(self, init_packet):
        def try_to_recover():
            if response['offset'] == 0 or response['offset'] > len(init_packet):
                # There is no init packet or present init packet is too long.
                return False

            expected_crc = (binascii.crc32(init_packet[:response['offset']]) & 0xFFFFFFFF)

            if expected_crc != response['crc']:
                # Present init packet is invalid.
                return False

            if len(init_packet) > response['offset']:
                # Send missing part.
                try:
                    self.__stream_data(data     = init_packet[response['offset']:],
                                       crc      = expected_crc,
                                       offset   = response['offset'])
                except ValidationException:
                    return False

            self.__execute()
            return True

        response = self.__select_command()
        assert len(init_packet) <= response['max_size'], 'Init command is too long'

        if try_to_recover():
            return

        try:
            self.__create_command(len(init_packet))
            self.__stream_data(data=init_packet)
            self.__execute()
        except ValidationException:
            raise NordicSemiException("Failed to send init packet")

    def send_firmware(self, firmware):
        def try_to_recover():
            if response['offset'] == 0:
                # Nothing to recover
                return

            expected_crc = binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
            remainder    = response['offset'] % response['max_size']

            if (expected_crc != response['crc']) or (remainder == 0):
                # Invalid CRC. Remove corrupted data.
                # DFU lib can't deal with recovering straight to an execute commamd,
                # so treat remainder of 0 as a corrupted block too.
                response['offset'] -= remainder if remainder != 0 else response['max_size']
                response['crc']     = \
                        binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
                return

            if (remainder != 0) and (response['offset'] != len(firmware)):
                # Send rest of the page.
                try:
                    to_send             = firmware[response['offset'] : response['offset']
                                                + response['max_size'] - remainder]
                    response['crc']     = self.__stream_data(data   = to_send,
                                                             crc    = response['crc'],
                                                             offset = response['offset'])
                    response['offset'] += len(to_send)
                except ValidationException:
                    # Remove corrupted data.
                    response['offset'] -= remainder
                    response['crc']     = \
                        binascii.crc32(firmware[:response['offset']]) & 0xFFFFFFFF
                    return

            self.__execute()
            self._send_event(event_type=DfuEvent.PROGRESS_EVENT, progress=response['offset'])

        response = self.__select_data()
        try_to_recover()
        for i in range(response['offset'], len(firmware), response['max_size']):
            data = firmware[i:i+response['max_size']]
            try:
                self.__create_data(len(data))
                response['crc'] = self.__stream_data(data=data, crc=response['crc'], offset=i)
                self.__execute()
            except ValidationException:
                raise NordicSemiException("Failed to send firmware")

            self._send_event(event_type=DfuEvent.PROGRESS_EVENT, progress=len(data))

    def __set_prn(self):
        logger.debug("ANT: Set Packet Receipt Notification {}".format(self.prn))
        self.dfu_adapter.send_message([DfuTransportAnt.OP_CODE['SetPRN']]
            + list(struct.pack('<H', self.prn)))
        self.__get_response(DfuTransportAnt.OP_CODE['SetPRN'])

    def __get_mtu(self):
        self.dfu_adapter.send_message([DfuTransportAnt.OP_CODE['GetSerialMTU']])
        response = self.__get_response(DfuTransportAnt.OP_CODE['GetSerialMTU'])

        self.mtu = struct.unpack('<H', bytearray(response))[0]

    def __ping(self):
        self.ping_id = (self.ping_id + 1) % 256

        self.dfu_adapter.send_message([DfuTransportAnt.OP_CODE['Ping'], self.ping_id])
        resp = self.dfu_adapter.get_message() # Receive raw response to check return code

        if (resp is None):
            logger.debug('ANT: No ping response')
            return False

        if resp[0] != DfuTransportAnt.OP_CODE['Response']:
            logger.debug('ANT: No Response: 0x{:02X}'.format(resp[0]))
            return False

        if resp[1] != DfuTransportAnt.OP_CODE['Ping']:
            logger.debug('ANT: Unexpected Executed OP_CODE.\n' \
                + 'Expected: 0x{:02X} Received: 0x{:02X}'.format(DfuTransportAnt.OP_CODE['Ping'], resp[1]))
            return False

        if resp[2] != DfuTransport.RES_CODE['Success']:
            # Returning an error code is seen as good enough. The bootloader is up and running
            return True
        else:
            if struct.unpack('B', bytearray(resp[3:]))[0] == self.ping_id:
                return True
            else:
                return False

    def __create_command(self, size):
        self.__create_object(0x01, size)

    def __create_data(self, size):
        self.__create_object(0x02, size)

    def __create_object(self, object_type, size):
        self.dfu_adapter.send_message([DfuTransportAnt.OP_CODE['CreateObject'], object_type]\
                                            + list(struct.pack('<L', size)))
        self.__get_response(DfuTransportAnt.OP_CODE['CreateObject'])

    def __calculate_checksum(self):
        self.dfu_adapter.send_message([DfuTransportAnt.OP_CODE['CalcChecSum']])
        response = self.__get_response(DfuTransportAnt.OP_CODE['CalcChecSum'])

        (offset, crc) = struct.unpack('<II', bytearray(response))
        return {'offset': offset, 'crc': crc}

    def __execute(self):
        self.dfu_adapter.send_message([DfuTransportAnt.OP_CODE['Execute']])
        self.__get_response(DfuTransportAnt.OP_CODE['Execute'])

    def __select_command(self):
        return self.__select_object(0x01)

    def __select_data(self):
        return self.__select_object(0x02)

    def __select_object(self, object_type):
        logger.debug("ANT: Selecting Object: type:{}".format(object_type))
        self.dfu_adapter.send_message([DfuTransportAnt.OP_CODE['ReadObject'], object_type])

        response = self.__get_response(DfuTransportAnt.OP_CODE['ReadObject'])
        (max_size, offset, crc)= struct.unpack('<III', bytearray(response))

        logger.debug("ANT: Object selected: " +
            " max_size:{} offset:{} crc:{}".format(max_size, offset, crc))
        return {'max_size': max_size, 'offset': offset, 'crc': crc}

    def __get_checksum_response(self):
        resp = self.__get_response(DfuTransportAnt.OP_CODE['CalcChecSum'])

        (offset, crc) = struct.unpack('<II', bytearray(resp))
        return {'offset': offset, 'crc': crc}

    def __stream_data(self, data, crc=0, offset=0):
        logger.debug("ANT: Streaming Data: " +
            "len:{0} offset:{1} crc:0x{2:08X}".format(len(data), offset, crc))
        def validate_crc():
            if (crc != response['crc']):
                raise ValidationException('Failed CRC validation.\n'\
                                + 'Expected: {} Received: {}.'.format(crc, response['crc']))
            if (offset != response['offset']):
                raise ValidationException('Failed offset validation.\n'\
                                + 'Expected: {} Received: {}.'.format(offset, response['offset']))

        current_pnr     = 0

        for i in range(0, len(data), self.mtu - 4):
            # append the write data opcode to the front
            # here the maximum data size is self.mtu - 4
            # due to the header bytes in commands.
            to_transmit = data[i:i + self.mtu - 4 ]
            to_transmit = struct.pack('B',DfuTransportAnt.OP_CODE['WriteObject']) + to_transmit

            self.dfu_adapter.send_message(list(to_transmit))
            crc     = binascii.crc32(to_transmit[1:], crc) & 0xFFFFFFFF
            offset += len(to_transmit) - 1
            current_pnr    += 1
            if self.prn == current_pnr:
                current_pnr = 0
                response    = self.__get_checksum_response()
                validate_crc()
        response = self.__calculate_checksum()
        validate_crc()
        return crc

    def __get_response(self, operation):
        def get_dict_key(dictionary, value):
            return next((key for key, val in list(dictionary.items()) if val == value), None)

        resp = self.dfu_adapter.get_message()

        if resp is None:
            return None

        if resp[0] != DfuTransportAnt.OP_CODE['Response']:
            raise NordicSemiException('No Response: 0x{:02X}'.format(resp[0]))

        if resp[1] != operation:
            raise NordicSemiException('Unexpected Executed OP_CODE.\n' \
                             + 'Expected: 0x{:02X} Received: 0x{:02X}'.format(operation, resp[1]))

        if resp[2] == DfuTransport.RES_CODE['Success']:
            return resp[3:]

        elif resp[2] == DfuTransport.RES_CODE['ExtendedError']:
            try:
                data = DfuTransport.EXT_ERROR_CODE[resp[3]]
            except IndexError:
                data = "Unsupported extended error type {}".format(resp[3])
            raise NordicSemiException('Extended Error 0x{:02X}: {}'.format(resp[3], data))
        else:
            raise NordicSemiException('Response Code {}'.format(
                get_dict_key(DfuTransport.RES_CODE, resp[2])))



#
# Copyright (c) 2018 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import unittest
import tempfile
import shutil
import os
import filecmp

from nordicsemi.zigbee.prod_config import ProductionConfig, ProductionConfigTooLargeException, ProductionConfigWrongException

class TestProductionConfig(unittest.TestCase):
    PRODUCTION_CONFIG_EXAMPLES_PATH = 'configs'
    PRODUCTION_CONFIG_GOLDEN_DATA_PATH = 'golden_data'

    def setUp(self):
        ''' Switch to a directory of the file and create a temporary directory to work in. '''
        script_abspath = os.path.abspath(__file__)
        script_dirname = os.path.dirname(script_abspath)
        os.chdir(script_dirname)

        self.work_directory = tempfile.mkdtemp(prefix="nrf_zigbee_production_config_tests_")

    def tearDown(self):
        ''' Remove a temporary directory. '''
        shutil.rmtree(self.work_directory, ignore_errors=True)

    def process_yaml_config(self, name):
        ''' Process the YAML config and generate an output hex file. '''
        input = os.path.join(self.PRODUCTION_CONFIG_EXAMPLES_PATH, name + '.yaml')
        output = os.path.join(self.work_directory, name + '.hex')

        try:
            pc = ProductionConfig(input)
        except ProductionConfigWrongException:
            self.fail("Error: Input YAML file format wrong.")

        try:
            pc.generate(output)
        except ProductionConfigTooLargeException as e:
            self.fail("Error: Production Config is too large.")

    def compare_hex_with_golden_data(self, name):
        ''' Compare the generated hex file with the golden vector. '''
        test = os.path.join(self.work_directory, name + '.hex')
        gold = os.path.join(self.PRODUCTION_CONFIG_GOLDEN_DATA_PATH, name + '.hex')

        return filecmp.cmp(test, gold)

    def generate_and_verify_production_config(self, name):
        ''' Generate and verify the production config. '''
        self.process_yaml_config(name)
        return self.compare_hex_with_golden_data(name)

    def test_channel_install_ieee_power_config(self):
        ''' Test config which contains 802.15.4 channel,
            install code, IEEE address and maximum Tx power.
        '''
        self.assertTrue(self.generate_and_verify_production_config('channel_install_ieee_power'))

    def test_install_ieee_power_config(self):
        ''' Test config which contains install code,
            IEEE address and maximum Tx power.
        '''
        self.assertTrue(self.generate_and_verify_production_config('install_ieee_power'))

    def test_install_ieee_config(self):
        self.assertTrue(self.generate_and_verify_production_config('install_ieee'))

    def test_install_config(self):
        self.assertTrue(self.generate_and_verify_production_config('install'))

    def test_empty_config(self):
        input = os.path.join(self.PRODUCTION_CONFIG_EXAMPLES_PATH, 'empty.yaml')
        self.assertRaises(ProductionConfigWrongException, ProductionConfig, input)

    def test_corrupt_config(self):
        input = os.path.join(self.PRODUCTION_CONFIG_EXAMPLES_PATH, 'corrupt.yaml')
        self.assertRaises(ProductionConfigWrongException, ProductionConfig, input)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import logging
import binascii
import struct
import tqdm
import threading

import piccata.core
import piccata.block_transfer
from piccata.block_transfer import extract_block
from piccata.message import Message
from piccata import constants
from ipaddress import ip_address
from collections import namedtuple
import click
import time
import math

logger = logging.getLogger(__name__)

def _make_trigger(init_data, image_data, mcast_mode = False, reset_suppress = 0):
    '''Create a trigger payload from given init and image data'''

    TRIGGER_VERSION    = 1
    MCAST_MODE_BIT     = 0x08
    RESET_SUPPRESS_BIT = 0x04

    def crc(payload):
        return binascii.crc32(payload) & 0xffffffff
    # Structure format:
    # flags:      uint8_t
    #
    #    |V3|V2|V1|V0|M|R|R1|R0|
    #
    #    V3-V0: version
    #    M:     mcast mode
    #    R:     reset suppress
    #    R1-R0: reserved bits
    #
    # init size:  uint32_t
    # init crc:   uint32_t
    # image size: uint32_t
    # image crc:  uint32_t
    flags = (TRIGGER_VERSION << 4)
    if (mcast_mode):
        flags |= MCAST_MODE_BIT
    if (reset_suppress != 0):
        flags |= RESET_SUPPRESS_BIT

    return struct.pack(">BIIII",
                       flags,
                       len(init_data),
                       crc(init_data),
                       len(image_data),
                       crc(image_data))

def _make_bitmap(resource):
    return [(resource, i) for i in range(0, _block_count(len(resource.data), ThreadDfuServer.BLOCK_SZX))]

def _block_count(length, block_size):
    '''Return number of blocks of a given size for the total length of data.'''
    return math.ceil(length / (2 ** (block_size + 4)))

def _bmp_to_str(bitmap):
    '''Convert binary data into a bit string'''
    s = ''
    for i in range(8):
        s = s + '{:08b} '.format((bitmap >> 64 - 8*(i + 1)) & 0xff)
    return s[:len(s) - 1]

def _get_block_opt(request):
    if (request.opt.block1):
        return request.opt.block1
    elif (request.opt.block2):
        return request.opt.block2
    else:
        return (0, False, constants.DEFAULT_BLOCK_SIZE_EXP)

class ThreadDfuClient:
    def __init__(self):
        '''Stores a reference to a progress bar object.'''
        self.progress_bar = None
        '''The number of a block most recently requested by a node.'''
        self.last_block = None

Resource = namedtuple('Resource', ['path', 'data'])

class ThreadDfuServer:
    REALM_LOCAL_ADDR  = ip_address('FF03::1')

    SPBLK_SIZE        = 64      # number of CoAP blocks of BLOCK_SZX size each
    SPBLK_UPLOAD_RATE = 1       # in blocks / seconds
    SPBLK_BMP_TIMEOUT = 2       # in seconds
    BLOCK_SZX         = 2       # 64 bytes
    ERASE_DELAY       = 0.5     # in seconds
    SPBLK_FLUSH_DELAY = 1.0     # delay between superblocks
    POST_UPLOAD_DELAY = 5.0     # delay after uploading the last block, in seconds

    IMAGE_URI = b'f'
    INIT_URI = b'i'
    TRIGGER_URI = b't'
    BITMAP_URI = b'b'

    def __init__(self, protocol, init_data, image_data, opts):
        assert(protocol is not None)
        assert(init_data is not None)
        assert(image_data is not None)

        self.opts = opts
        if (not opts or not opts.rate):
            self.opts.rate = self.SPBLK_UPLOAD_RATE
        if (not opts or not opts.mcast_dfu):
            self.opts.mcast_dfu = False
        if (not opts or not opts.reset_suppress):
            self.opts.reset_suppress = 0

        self.protocol = protocol
        self.protocol.register_request_handler(self)

        self.progress_bar = None

        self.missing_blocks = []
        self.bmp_received_event = threading.Event()
        self.upload_done_event = threading.Event()
        self.upload_done_event.set()
        self.trig_done_event = threading.Event()

        self.init_resource = Resource((ThreadDfuServer.INIT_URI,), init_data)
        self.image_resource = Resource((ThreadDfuServer.IMAGE_URI,), image_data)

        self.clients = {}
        self.upload_thread = None

    def _draw_token(self):
        return piccata.message.random_token(2)

    def _update_progress_bar(self, address, client, block_count, total_block_count):
        # If node didn't request any blocks yet then create a new progress
        # bar for it. Update otherwise.
        if (client.progress_bar is None):
            client.progress_bar = tqdm.tqdm(desc = str(address),
                                            position = len(self.clients) - 1,
                                            initial = block_count,
                                            total = total_block_count)
        elif (block_count > client.last_block):
            client.progress_bar.update(block_count - client.last_block)

        if (block_count == total_block_count - 1):
            # One last update to fill the progress bar (block_count is indexed from 0)
            client.progress_bar.update()
            client.progress_bar.close()
            client.progress_bar = None

    def _handle_image_request(self, request):
        if (request.remote not in self.clients):
            self.clients[request.remote] = ThreadDfuClient()

        block_num, _, block_szx = _get_block_opt(request)

        total_block_count = _block_count(len(self.image_resource.data), block_szx)

        self._update_progress_bar(request.remote.addr,
                                  self.clients[request.remote],
                                  block_num,
                                  total_block_count)

        if (self.clients[request.remote].last_block is None) or (self.clients[request.remote].last_block < block_num):
                self.clients[request.remote].last_block = block_num

        if block_num == total_block_count - 1:
            self.clients[request.remote].last_block = None
            click.echo() # New line after progress bar
            click.echo("Thread DFU upload complete")

        return piccata.block_transfer.create_block_2_response(self.image_resource.data, request)

    def _handle_init_request(self, request):
        # Add remote to the list of prospective DFU clients
        if (request.remote not in self.clients):
            self.clients[request.remote] = ThreadDfuClient()
            logger.debug("Added {} to clients".format(request.remote.addr))

        return piccata.block_transfer.create_block_2_response(self.init_resource.data, request)

    def _handle_trigger_response(self, result, request, response, num_of_requests):
        assert (result == piccata.constants.RESULT_TIMEOUT)

        if (num_of_requests - 1 > 0):
            self.protocol.request(request,
                                  self._handle_trigger_response,
                                  (num_of_requests - 1,))
            return

        self.trig_done_event.set()

    def _handle_trigger_request(self, request):
        response = None
        if request.mtype == piccata.constants.CON:
            response = Message.AckMessage(request,
                                          constants.CONTENT,
                                          _make_trigger(self.init_resource.data,
                                                        self.image_resource.data,
                                                        False,
                                                        self.opts.reset_suppress))
        else:
            if self.opts.mcast_dfu:
                address = self.REALM_LOCAL_ADDR
            else:
                address = request.remote.addr

            self.trigger(address, 3)

        return response

    def _send_block(self, remote, path, num, more, szx, payload):
        '''
        Send a single block.
        :param remote: An address of the remote endpoint.
        :param path: An URI path to the block resource.
        :param num: A block number. Part of the CoAP block option.
        :param more: An information if more blocks are pending. Part of the CoAP block option.
        :param szx: A block size, encoded in CoAP block option format. Part of the CoAP block option.
        :param payload: A block payload.
        '''
        logger.info('Sending block {} to {}'.format(num, remote.addr))

        request = piccata.message.Message(mtype = piccata.constants.NON,
                                          code = piccata.constants.PUT)

        request.opt.uri_path = path
        request.opt.block1 = (num, more, szx)
        request.remote = remote
        request.payload = payload

        self.protocol.request(request)

    def _upload(self, remote, bitmap):
        while True:
            # Bitmap holds (resource, num) tuples. Sort them using path and block num.
            bitmap.sort(key = lambda item : (item[0].path[0] == ThreadDfuServer.IMAGE_URI, item[1]))
            resource, num = bitmap.pop(0)

            payload, more = extract_block(resource.data,
                                          num,
                                          ThreadDfuServer.BLOCK_SZX)

            logger.debug("Uploading resource {} block {} to {}".format(resource.path, num, remote.addr))

            total_block_count = _block_count(len(resource.data), ThreadDfuServer.BLOCK_SZX)

            self._update_progress_bar(remote.addr,
                                      self.clients[remote],
                                      num,
                                      total_block_count)

            self._send_block(remote,
                             resource.path,
                             num,
                             more,
                             ThreadDfuServer.BLOCK_SZX,
                             payload)

            if (self.clients[remote].last_block is None) or (self.clients[remote].last_block < num):
                self.clients[remote].last_block = num

            if num == total_block_count - 1:
                self.clients[remote].last_block = None

            self.bmp_received_event.clear()
            if len(bitmap):
                if (num % ThreadDfuServer.SPBLK_SIZE == 0) or (((num + 1) % ThreadDfuServer.SPBLK_SIZE) == 0):
                    delay = ThreadDfuServer.ERASE_DELAY
                else:
                    delay = 1.0/self.opts.rate

                time.sleep(delay)

            else:
                self.upload_done_event.set()
                self.bmp_received_event.wait()

    def _handle_reset_response(self, result, request, response, num_of_requests, delay):
        assert (result == piccata.constants.RESULT_TIMEOUT)

        if (num_of_requests - 1 > 0):
            self.protocol.request(request,
                                  self._handle_reset_response,
                                  (num_of_requests - 1, delay))

    def _send_reset_request(self, remote, num_of_requests, delay):
        logger.info('Sending reset request to {}'.format(remote))

        request = piccata.message.Message(mtype = piccata.constants.NON,
                                          code = piccata.constants.PUT,
                                          token = self._draw_token())

        request.opt.uri_path = (b"r",)
        request.remote = remote
        request.timeout = ThreadDfuServer.SPBLK_BMP_TIMEOUT
        request.payload = struct.pack(">I", delay)

        self.protocol.request(request,
                              self._handle_reset_response,
                              (num_of_requests, delay))

    def _handle_bitmap_request(self, request):
        payload = struct.unpack('!HQ', request.payload)
        num = payload[0]
        bmp = payload[1]
        path = request.opt.uri_path[1]
        logger.debug("Device {} returned path {} num {} bmp {}".format(request.remote.addr,
                                                                       path,
                                                                       num,
                                                                       _bmp_to_str(bmp)))

        if (path == ThreadDfuServer.INIT_URI):
            resource = self.init_resource
        elif (path == ThreadDfuServer.IMAGE_URI):
            resource = self.image_resource

        for i in range(ThreadDfuServer.SPBLK_SIZE):
            item = (resource, num + i)
            if (bmp & (1 << (ThreadDfuServer.SPBLK_SIZE - 1 - i)) and item not in self.missing_blocks):
                self.missing_blocks.append(item)
                logger.debug("Added {} block {} to missing list".format(item[0].path, item[1]))

        self.bmp_received_event.set()
        return None

    def receive_request(self, request):
        '''Request callback called by the CoAP toolkit. Note that the function
           signature (name, args) is expected by the CoAP toolkit.'''

        # TODO: consider a case where there is a mcast DFU in progress
        #       but a request is received. How this should be handled?

        handlers = {
            ThreadDfuServer.IMAGE_URI : self._handle_image_request,
            ThreadDfuServer.INIT_URI : self._handle_init_request,
            ThreadDfuServer.TRIGGER_URI : self._handle_trigger_request,
            ThreadDfuServer.BITMAP_URI : self._handle_bitmap_request,
        }

        for uri, handler in list(handlers.items()):
            if b'/'.join(request.opt.uri_path).startswith(uri):
                return handler(request)

        return piccata.message.Message.AckMessage(request, piccata.constants.NOT_FOUND)

    def _multicast_upload(self, remote, num_of_requests):
        self.upload_done_event.clear()

        click.echo("Waiting 20s before starting multicast DFU procedure")
        time.sleep(20)

        self.missing_blocks.extend(_make_bitmap(self.init_resource))
        self.missing_blocks.extend(_make_bitmap(self.image_resource))

        self.clients[remote] = ThreadDfuClient()

        self.bmp_received_event.set()

        self._send_trigger(remote, num_of_requests)
        self.trig_done_event.wait()

        if self.upload_thread is None or self.upload_thread.is_alive() is False:
            self.upload_thread = threading.Thread(target = self._upload,
                                                  name = "Upload thread",
                                                  args = (remote, self.missing_blocks))
            self.upload_thread.setDaemon(True)
            self.upload_thread.start()

        time.sleep(15)
        self.upload_done_event.wait()

        if (self.opts.reset_suppress > 0):
            self._send_reset_request(remote, num_of_requests, self.opts.reset_suppress)

        click.echo() # New line after progress bar
        click.echo("Thread DFU upload complete")

    def _send_trigger(self, remote, num_of_requests):
        logger.info('Triggering DFU on {}'.format(remote))
        request = piccata.message.Message(mtype = piccata.constants.NON,
                                          code = piccata.constants.POST,
                                          token = self._draw_token())

        request.opt.uri_path = (ThreadDfuServer.TRIGGER_URI,)
        request.remote = remote
        request.timeout = ThreadDfuServer.SPBLK_BMP_TIMEOUT
        request.payload = _make_trigger(self.init_resource.data,
                                        self.image_resource.data,
                                        self.opts.mcast_dfu,
                                        self.opts.reset_suppress)
        self.protocol.request(request,
                              self._handle_trigger_response,
                              (num_of_requests,))

    def trigger(self, address, num_of_requests):
        remote = piccata.types.Endpoint(address, piccata.constants.COAP_PORT)

        if self.opts.mcast_dfu:
            if self.upload_done_event.is_set():
                thread = threading.Thread(target = self._multicast_upload,
                                          args = (remote, num_of_requests, ))
                thread.setDaemon(True)
                thread.start()
        else:
            self._send_trigger(remote, num_of_requests)



#
# Copyright (c) 2018 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import struct

OTA_UPGRADE_FILE_HEADER_FILE_ID       = 0x0BEEF11E
OTA_UPGRADE_FILE_HEADER_FILE_VERSION  = 0x0100
OTA_UPGRADE_FILE_HEADER_LENGTH        = 4 + 2 + 2 + 2 + 2 + 2 + 4 + 2 + 32 + 4
OTA_UPGRADE_FIELD_CONTROL             = 0x00
OTA_UPGRADE_MANUFACTURER_WILDCARD     = 0xFFFF
OTA_UPGRADE_IMAGE_TYPE_WILDCARD       = 0xFFFF
OTA_UPGRADE_FILE_HEADER_STACK_PRO     = 0x0002

OTA_UPGRADE_SUBELEMENT_HEADER_SIZE    = 2 + 4  # Tag ID + Length Field
OTA_UPGRADE_SUBELEMENT_TRIGGER_TYPE   = 0xCDEF
OTA_UPGRADE_SUBELEMENT_TRIGGER_LENGTH = 1 + 4 + 4 + 4 + 4 # See background_dfu_trigger_t in the nRF5 SDK for explanation

OTA_UPGRADE_FIELD_CONTROL_BIT_MASK_HW_VER = (1 << 2)
OTA_UPGRADE_MIN_HW_VERSION_LENGTH         = 2
OTA_UPGRADE_MAX_HW_VERSION_LENGTH         = 2
'''
   background_dfu_trigger_t type defined in components/iot/background_dfu/background_dfu_state.h of nRF5 SDK:

    /** @brief Trigger packet structure. */
    typedef PACKED_STRUCT
    {
        uint8_t  flags;         /**< Trigger message flags. Bits 7:4 (oldest) - trigger version, bit 3 - DFU mode, bits 2:0 - reserved. */
        uint32_t init_length;
        uint32_t init_crc;
        uint32_t image_length;
        uint32_t image_crc;
    } background_dfu_trigger_t;
'''

class OTA_file:

    def __init__(self,
                 file_version,
                 init_cmd_len,
                 init_cmd_crc,
                 init_cmd,
                 firmware_len,
                 firmware_crc,
                 firmware,
                 manufacturer_code = OTA_UPGRADE_MANUFACTURER_WILDCARD,
                 image_type = OTA_UPGRADE_IMAGE_TYPE_WILDCARD,
                 comment = '',
                 min_hw_version=None,
                 max_hw_version=None):
        '''A constructor for the OTA file class, see Zigbee ZCL spec 11.4.2 (Zigbee Document 07-5123-06)
           see: http://www.zigbee.org/~zigbeeor/wp-content/uploads/2014/10/07-5123-06-zigbee-cluster-library-specification.pdf
           (access verified as of 2018-08-06)
        '''

        total_len = OTA_UPGRADE_FILE_HEADER_LENGTH + 3 * OTA_UPGRADE_SUBELEMENT_HEADER_SIZE + OTA_UPGRADE_SUBELEMENT_TRIGGER_LENGTH + init_cmd_len + firmware_len

        ota_header = OTA_header(OTA_UPGRADE_FILE_HEADER_FILE_ID,
                                OTA_UPGRADE_FILE_HEADER_FILE_VERSION,
                                OTA_UPGRADE_FILE_HEADER_LENGTH,
                                OTA_UPGRADE_FIELD_CONTROL,
                                manufacturer_code,
                                image_type,
                                file_version,
                                OTA_UPGRADE_FILE_HEADER_STACK_PRO,
                                comment,
                                total_len,
                                min_hw_version,
                                max_hw_version)

        subelement_header_pack_format = '<HL'
        subelement_trigger_pack_format = '>cLLLL'

        subelement_trigger = struct.pack(subelement_header_pack_format,
                                         OTA_UPGRADE_SUBELEMENT_TRIGGER_TYPE,
                                         OTA_UPGRADE_SUBELEMENT_TRIGGER_LENGTH)
        subelement_trigger_payload  = struct.pack(subelement_trigger_pack_format,
                                                  bytes([0x10]),  # flags
                                                  init_cmd_len,
                                                  init_cmd_crc,
                                                  firmware_len,
                                                  firmware_crc)

        subelement_init_cmd = struct.pack(subelement_header_pack_format, 0x0000, init_cmd_len) # Subelement tags are not really needed in case of Init Cmd and Firmware subelements
        subelement_firmware = struct.pack(subelement_header_pack_format, 0x0000, firmware_len)

        self.binary = ota_header.header + subelement_trigger + subelement_trigger_payload + subelement_init_cmd + init_cmd + subelement_firmware + firmware
        self.filename = '-'.join(['{:04X}'.format(manufacturer_code),
                                  '{:04X}'.format(image_type),
                                  '{:08X}'.format(file_version),
                                  comment]) + '.zigbee'


class OTA_header:
    def __init__(self,
                 file_id,
                 header_version,
                 header_length,
                 header_field_control,
                 manufacturer_code,
                 image_type,
                 file_version,
                 zigbee_stack_version,
                 header_string,
                 total_image_size,
                 min_hw_version=None,
                 max_hw_version=None):

        self.__pack_format = '<LHHHHHLHc31sL'
        self.__header_length = header_length
        self.__total_size = total_image_size
        self.__field_control = header_field_control
        self.__additional_fields = []

        # If min and max hardware version are present add optional fields to the header
        if (type(min_hw_version) is int) and (type(max_hw_version) is int):
            self.__add_optional_fields((OTA_UPGRADE_MIN_HW_VERSION_LENGTH + OTA_UPGRADE_MAX_HW_VERSION_LENGTH),
                                       OTA_UPGRADE_FIELD_CONTROL_BIT_MASK_HW_VER,
                                       'HH',
                                       [min_hw_version, max_hw_version])

        self.__pack_args = [file_id,
                            header_version,
                            self.__header_length,
                            self.__field_control,
                            manufacturer_code,
                            image_type,
                            file_version,
                            zigbee_stack_version,
                            bytes([len(header_string)]),
                            bytes(header_string.encode('ascii')),
                            self.__total_size]
        self.__pack_args.extend(self.__additional_fields)

    @property
    def header(self):
        return struct.pack(self.__pack_format, *self.__pack_args)

    def __add_optional_fields(self, fields_length, field_control_bit_mask, fields_formatting, fields_values):
        self.__total_size    += fields_length
        self.__header_length += fields_length
        self.__field_control += field_control_bit_mask
        self.__pack_format   += fields_formatting
        if type(fields_values) in (tuple, list):
            self.__additional_fields.extend(fields_values)
        else:
            self.__additional_fields.append(fields_values)



#
# Copyright (c) 2016 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import re
import subprocess
from pc_ble_driver_py.ble_driver import Flasher

class NCPFlasher(Flasher):
    ERROR_CODE_VERIFY_ERROR = 55

    def __init__(self, serial_port = None, snr = None):
        super().__init__(serial_port, snr)

    def __get_hex_path(self):
        return os.path.join(os.path.dirname(__file__), 'hex', 'ncp.hex')

    def verify(self, path):
        args = ['--verify', path]
        return self.call_cmd(args)

    def fw_check(self):
        try:
            result = self.verify(self.__get_hex_path())
        except subprocess.CalledProcessError as e:
            if (e.returncode == NCPFlasher.ERROR_CODE_VERIFY_ERROR):
                return False
            else:
                raise
        return (re.search(b'^Verified OK.$', result, re.MULTILINE) is not None)

    def fw_flash(self):
        self.erase()
        self.program(self.__get_hex_path())



#
# Copyright (c) 2018 Nordic Semiconductor ASA
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this
#   list of conditions and the following disclaimer in the documentation and/or
#   other materials provided with the distribution.
#
#   3. Neither the name of Nordic Semiconductor ASA nor the names of other
#   contributors to this software may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   4. This software must only be used in or with a processor manufactured by Nordic
#   Semiconductor ASA, or in or with a processor manufactured by a third party that
#   is used in combination with a processor manufactured by Nordic Semiconductor.
#
#   5. Any software provided in binary or object form under this license must not be
#   reverse engineered, decompiled, modified and/or disassembled.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import os
import re
import subprocess
import time
import uuid
from intelhex import IntelHex
from serial import Serial
from pc_ble_driver_py.ble_driver import Flasher

class OTAFlasher(Flasher):
    ERROR_CODE_VERIFY_ERROR = 55
    # Address in flash memory, where the Zigbee image will be placed. This value has to be synced with UPGRADE_IMAGE_OFFSET constant inside Zigbee OTA server source code.
    OTA_UPDATE_OFFSET = 0x80000
    OTA_EUI64_PREFIX = '07A07A'

    def __init__(self, fw, channel = 16, serial_port = None, snr = None):
        '''
           Ininitialises the OTA Flasher class which handles flashing the devboard
           with the needed OTA Server firmware and the update file for the OTA Client.
           The said devboard shall become the OTA Server which shall propagate the
           image on the Zigbee network.

           Keyword arguments:
           fw -- path to the update file for the OTA Client
           channel -- a 802.15.4 channel number, on which the OTA Server shall operate (default 16)
           serial_port -- a serial port of the connected devboard which shall be flashed with OTA Server
           snr -- a JLink serial number of the connected devboard which shall be flashed with OTA Server

           Note: only one parameter out of (serial_port, snr) must be provided, since the superclass
                 constructor shall handle resolving the rest.

        '''
        # Call the superclass constructor
        super().__init__(serial_port, snr)
        # Create a Intel Hex out of the Zigbee Update file
        ih = IntelHex()
        update = open(fw, 'rb').read()
        ih.puts(OTAFlasher.OTA_UPDATE_OFFSET, update)
        self.update_firmware_hex = fw + '.hex'
        ih.write_hex_file(self.update_firmware_hex)
        # Open the serial channel to the devboard and save the 802.15.4 channel
        self.ser = Serial(self.serial_port, 115200)
        self.channel = channel

    def __get_hex_path(self):
        '''Return the absolute path to the firmware of the OTA Server'''
        return os.path.join(os.path.dirname(__file__), 'hex', 'ota.hex')

    def verify(self, path):
        '''Run the verify command'''
        args = ['--verify', path]
        return self.call_cmd(args)

    def _fw_check(self, path_to_file):
        '''Check if the path_to_file hexfile was flashed correctly'''
        try:
            result = self.verify(path_to_file)
        except subprocess.CalledProcessError as e:  # for pc-ble-driver <= 0.14.2, can be removed when requirements
            # will be updated to >= 0.15.0
            if e.returncode == OTAFlasher.ERROR_CODE_VERIFY_ERROR:
                return False
            else:
                raise
        except RuntimeError:  # for pc-ble-driver >= 0.15.0
            return False

        return (re.search(b'^Verified OK.$', result, re.MULTILINE) is not None)

    def fw_check(self):
        '''Check if all the hex files (OTA Server firmware and Zigbee Update file) were flashed correctly'''
        if self._fw_check(self.__get_hex_path()) and self._fw_check(self.update_firmware_hex):
            return True
        else:
            return False

    def fw_flash(self):
        '''Flash all the hex files (OTA Server firmware and Zigbee Update file) to the board'''
        self.erase()
        self.program(self.update_firmware_hex)
        self.program(self.__get_hex_path())
        # Remove the generated file
        os.remove(self.update_firmware_hex)

    def randomize_eui64(self):
        '''Randomize the EUI64 address used by the board'''
        random_eui64 = uuid.uuid4().int >> 88 # Generate 128-bit UUID and take 40 upper bits
        self.ser.write(f'zdo eui64 {OTAFlasher.OTA_EUI64_PREFIX}{random_eui64:010x}\r\n'.encode())

    def setup_channel(self):
        '''Setup to the channel of the flashed board through the serial CLI; and start the internal stack'''
        self.ser.write(f'bdb channel {self.channel}\r\n'.encode())
        time.sleep(1.0)
        self.ser.write('bdb start\r\n'.encode())
